import logging
import os
from typing import (
    Optional,
)

from fastapi import (
    Body,
    Request,
    Response,
    status,
)
from pydantic import BaseModel
from sqlalchemy import (
    false,
    true,
    update,
)

import tool_shed.util.shed_util_common as suc
from galaxy.exceptions import (
    InsufficientPermissionsException,
    ObjectNotFound,
    RequestParameterInvalidException,
)
from galaxy.managers.api_keys import ApiKeyManager
from galaxy.managers.users import UserManager
from galaxy.webapps.base.webapp import create_new_session
from tool_shed.context import SessionRequestContext
from tool_shed.managers.users import (
    api_create_user,
    get_api_user,
    index,
)
from tool_shed.structured_app import ToolShedApp
from tool_shed.webapp.fast_app import limiter
from tool_shed.webapp.model import (
    GalaxySession,
    User as SaUser,
)
from tool_shed_client.schema import (
    CreateUserRequest,
    UserV2 as User,
)
from . import (
    depends,
    DependsOnTrans,
    ensure_valid_session,
    Router,
    set_auth_cookie,
    UserIdPathParam,
)

router = Router(tags=["users"])

log = logging.getLogger(__name__)

TOOL_SHED_SENSITIVE_API_REQUEST_LIMIT: Optional[str] = os.environ.get("TOOL_SHED_SENSITIVE_API_REQUEST_LIMIT", None)
SENSITIVE_API_REQUEST_LIMIT = TOOL_SHED_SENSITIVE_API_REQUEST_LIMIT or "10/minute"


class UiRegisterRequest(BaseModel):
    email: str
    username: str
    password: str
    bear_field: str


class HasCsrfToken(BaseModel):
    session_csrf_token: str


class UiLoginRequest(HasCsrfToken):
    login: str
    password: str


class UiLogoutRequest(HasCsrfToken):
    logout_all: bool = False


class UiLoginResponse(BaseModel):
    pass


class UiLogoutResponse(BaseModel):
    pass


class UiRegisterResponse(BaseModel):
    email: str
    activation_sent: bool = False
    activation_error: bool = False
    contact_email: Optional[str] = None


class UiChangePasswordRequest(BaseModel):
    current: str
    password: str


INVALID_LOGIN_OR_PASSWORD = "Invalid login or password"


@router.cbv
class FastAPIUsers:
    app: ToolShedApp = depends(ToolShedApp)
    user_manager: UserManager = depends(UserManager)
    api_key_manager: ApiKeyManager = depends(ApiKeyManager)

    @router.get(
        "/api/users",
        description="index users",
        operation_id="users__index",
    )
    def index(self, trans: SessionRequestContext = DependsOnTrans) -> list[User]:
        deleted = False
        return index(trans.app, deleted)

    @router.post(
        "/api/users",
        description="create a user",
        operation_id="users__create",
        require_admin=True,
    )
    def create(self, trans: SessionRequestContext = DependsOnTrans, request: CreateUserRequest = Body(...)) -> User:
        return api_create_user(trans, request)

    @router.get(
        "/api/users/current",
        description="show current user",
        operation_id="users__current",
    )
    def current(self, trans: SessionRequestContext = DependsOnTrans) -> User:
        user = trans.user
        if not user:
            raise ObjectNotFound()

        return get_api_user(trans.app, user)

    @router.get(
        "/api/users/{encoded_user_id}",
        description="show a user",
        operation_id="users__show",
    )
    def show(self, trans: SessionRequestContext = DependsOnTrans, encoded_user_id: str = UserIdPathParam) -> User:
        user = suc.get_user(trans.app, encoded_user_id)
        if user is None:
            raise ObjectNotFound()
        return get_api_user(trans.app, user)

    @router.get(
        "/api/users/{encoded_user_id}/api_key",
        name="get_or_create_api_key",
        summary="Return the user's API key",
        operation_id="users__get_or_create_api_key",
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def get_or_create_api_key(
        self, request: Request, trans: SessionRequestContext = DependsOnTrans, encoded_user_id: str = UserIdPathParam
    ) -> str:
        user = self._get_user(trans, encoded_user_id)
        return self.api_key_manager.get_or_create_api_key(user)

    @router.post(
        "/api/users/{encoded_user_id}/api_key",
        summary="Creates a new API key for the user",
        operation_id="users__create_api_key",
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def create_api_key(
        self, request: Request, trans: SessionRequestContext = DependsOnTrans, encoded_user_id: str = UserIdPathParam
    ) -> str:
        user = self._get_user(trans, encoded_user_id)
        return self.api_key_manager.create_api_key(user).key

    @router.delete(
        "/api/users/{encoded_user_id}/api_key",
        summary="Delete the current API key of the user",
        status_code=status.HTTP_204_NO_CONTENT,
        operation_id="users__delete_api_key",
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def delete_api_key(
        self,
        request: Request,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_user_id: str = UserIdPathParam,
    ):
        user = self._get_user(trans, encoded_user_id)
        self.api_key_manager.delete_api_key(user)
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    def _get_user(self, trans: SessionRequestContext, encoded_user_id: str):
        if encoded_user_id == "current":
            user = trans.user
        else:
            user = suc.get_user(trans.app, encoded_user_id)
        if user is None:
            raise ObjectNotFound()
        if not (trans.user_is_admin or trans.user == user):
            raise InsufficientPermissionsException()
        return user

    @router.post(
        "/api_internal/register",
        description="register a user",
        operation_id="users__internal_register",
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def register(
        self,
        request: Request,
        trans: SessionRequestContext = DependsOnTrans,
        register_request: UiRegisterRequest = Body(...),
    ) -> UiRegisterResponse:
        honeypot_field = register_request.bear_field
        if honeypot_field != "":
            message = "You've been flagged as a possible bot. If you are not, please try registering again and fill the form out carefully."
            raise RequestParameterInvalidException(message)

        username = register_request.username
        if username == "repos":
            raise RequestParameterInvalidException("Cannot create a user with the username 'repos'")
        self.user_manager.create(email=register_request.email, username=username, password=register_request.password)
        if self.app.config.user_activation_on:
            is_activation_sent = self.user_manager.send_activation_email(trans, register_request.email, username)
            if is_activation_sent:
                return UiRegisterResponse(email=register_request.email, activation_sent=True)
            else:
                return UiRegisterResponse(
                    email=register_request.email,
                    activation_sent=False,
                    activation_error=True,
                    contact_email=self.app.config.error_email_to,
                )
        else:
            return UiRegisterResponse(email=register_request.email)

    @router.put(
        "/api_internal/change_password",
        description="reset a user",
        operation_id="users__internal_change_password",
        status_code=status.HTTP_204_NO_CONTENT,
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def change_password(
        self,
        request: Request,
        trans: SessionRequestContext = DependsOnTrans,
        change_request: UiChangePasswordRequest = Body(...),
    ):
        password = change_request.password
        current = change_request.current
        if trans.user is None:
            raise InsufficientPermissionsException("Must be logged into use this functionality")
        user_id = trans.user.id
        token = None
        user, message = self.user_manager.change_password(
            trans, password=password, current=current, token=token, confirm=password, id=user_id
        )
        if not user:
            raise RequestParameterInvalidException(message)
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.put(
        "/api_internal/login",
        description="login to web UI",
        operation_id="users__internal_login",
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def internal_login(
        self, request: Request, trans: SessionRequestContext = DependsOnTrans, login_request: UiLoginRequest = Body(...)
    ) -> UiLoginResponse:
        log.info(f"top of internal_login {trans.session_csrf_token}")
        ensure_csrf_token(trans, login_request)
        login = login_request.login
        password = login_request.password
        user = self.user_manager.get_user_by_identity(login)
        if user is None:
            raise InsufficientPermissionsException(INVALID_LOGIN_OR_PASSWORD)
        elif user.deleted:
            message = (
                "This account has been marked deleted, contact your local Galaxy administrator to restore the account."
            )
            if trans.app.config.error_email_to is not None:
                message += f" Contact: {trans.app.config.error_email_to}."
            raise InsufficientPermissionsException(message)
        elif not trans.app.auth_manager.check_password(user, password, trans.request):
            raise InsufficientPermissionsException(INVALID_LOGIN_OR_PASSWORD)
        else:
            handle_user_login(trans, user)
        return UiLoginResponse()

    @router.put(
        "/api_internal/logout",
        description="logout of web UI",
        operation_id="users__internal_logout",
    )
    @limiter.limit(SENSITIVE_API_REQUEST_LIMIT)
    def internal_logout(
        self,
        request: Request,
        trans: SessionRequestContext = DependsOnTrans,
        logout_request: UiLogoutRequest = Body(...),
    ) -> UiLogoutResponse:
        ensure_csrf_token(trans, logout_request)
        handle_user_logout(trans, logout_all=logout_request.logout_all)
        return UiLogoutResponse()


def ensure_csrf_token(trans: SessionRequestContext, request: HasCsrfToken):
    session_csrf_token = request.session_csrf_token
    if not trans.session_csrf_token:
        ensure_valid_session(trans)
    message = None
    if not session_csrf_token:
        message = "No session token set, denying request."
    elif session_csrf_token != trans.session_csrf_token:
        log.info(f"{session_csrf_token} != {trans.session_csrf_token}")
        message = "Wrong session token found, denying request."
    if message:
        raise InsufficientPermissionsException(message)


def handle_user_login(trans: SessionRequestContext, user: SaUser) -> None:
    trans.app.security_agent.create_user_role(user, trans.app)
    replace_previous_session(trans, user)


def handle_user_logout(trans, logout_all=False):
    """
    Logout the current user:
        - invalidate current session + previous sessions (optional)
        - create a new session with no user associated
    """
    if logout_all:
        prev_session = trans.get_galaxy_session()
        if prev_session and prev_session.user_id:
            invalidate_user_sessions(trans.sa_session, prev_session.user_id)
    replace_previous_session(trans, None)


def replace_previous_session(trans, user):
    prev_galaxy_session = trans.get_galaxy_session()
    # Invalidate previous session
    if prev_galaxy_session:
        prev_galaxy_session.is_valid = False
    # Create new session
    new_session = create_new_session(trans, prev_galaxy_session, user)
    trans.set_galaxy_session(new_session)
    trans.sa_session.add_all((prev_galaxy_session, new_session))
    trans.sa_session.commit()
    set_auth_cookie(trans, new_session)


def invalidate_user_sessions(session, user_id):
    stmt = (
        update(GalaxySession)
        .values(is_valid=false())
        .where(GalaxySession.user_id == user_id)
        .where(GalaxySession.is_valid == true())
    )
    session.execute(stmt)
    session.commit()



import pytest

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_maker_repository_name = "column_maker_0020"
column_maker_repository_description = "A flexible aligner."
column_maker_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

emboss_repository_description = "Galaxy wrappers for Emboss version 5.0.0 tools"
emboss_repository_long_description = "Galaxy wrappers for Emboss version 5.0.0 tools"

emboss_repository_name = "emboss_0050"
emboss_5_repository_name = "emboss_5_0050"
emboss_6_repository_name = "emboss_6_0050"

filtering_repository_name = "filtering_0050"
filtering_repository_description = "Galaxy's filtering tool"
filtering_repository_long_description = "Long description of Galaxy's filtering tool"

freebayes_repository_name = "freebayes_0050"
freebayes_repository_description = "Galaxy's freebayes tool"
freebayes_repository_long_description = "Long description of Galaxy's freebayes tool"

column_repository_name = "column_maker_0050"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_0050"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

bismark_repository_name = "bismark_0050"
bismark_repository_description = "A flexible aligner."
bismark_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

category_0050_name = "Test 0050 Circular Dependencies 5 Levels"
category_0050_description = "Test circular dependency features"

running_standalone = False


class TestResetAllRepositoryMetadata(ShedTwillTestCase):
    """Verify that the "Reset selected metadata" feature works."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_filtering_repository(self):
        """Create and populate the filtering_0000 repository."""
        global running_standalone
        self.login(email=common.admin_email, username=common.admin_username)
        category_0000 = self.create_category(
            name="Test 0000 Basic Repository Features 1", description="Test 0000 Basic Repository Features 1"
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name="filtering_0000",
            description="Galaxy's filtering tool",
            long_description="Long description of Galaxy's filtering tool",
            owner=common.test_user_1_name,
            category=category_0000,
        )
        if self.repository_is_new(repository):
            running_standalone = True
            self.commit_tar_to_repository(
                repository,
                "filtering/filtering_1.1.0.tar",
                commit_message="Uploaded filtering 1.1.0 tarball.",
            )
            self.add_tar_to_repository(repository, "filtering/filtering_2.2.0.tar")

    def test_0010_create_freebayes_repository(self):
        """Create and populate the freebayes_0010 repository."""
        self.login(email=common.admin_email, username=common.admin_username)
        category_0010 = self.create_category(
            name="Test 0010 Repository With Tool Dependencies",
            description="Tests for a repository with tool dependencies.",
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name="freebayes_0010",
            description="Galaxy's freebayes tool",
            long_description="Long description of Galaxy's freebayes tool",
            owner=common.test_user_1_name,
            category=category_0010,
            strings_displayed=[],
        )
        if running_standalone:
            self.setup_freebayes_0010_repo(repository)

    def test_0015_create_datatypes_0020_repository(self):
        """Create and populate the column_maker_0020 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0020 = self.create_category(
                name="Test 0020 Basic Repository Dependencies",
                description="Testing basic repository dependency features.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name="column_maker_0020",
                description=column_maker_repository_description,
                long_description=column_maker_repository_long_description,
                owner=common.test_user_1_name,
                category=category_0020,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0020_create_emboss_0020_repository(self):
        """Create and populate the emboss_0020 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0020 = self.create_category(
                name="Test 0020 Basic Repository Dependencies",
                description="Testing basic repository dependency features.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name="emboss_0020",
                description=emboss_repository_long_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category_0020,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss.tar",
            )

    def test_0025_create_emboss_datatypes_0030_repository(self):
        """Create and populate the emboss_0030 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0030 = self.create_category(
                name="Test 0030 Repository Dependency Revisions",
                description="Testing repository dependencies by revision.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            column_maker_repository = self.get_or_create_repository(
                name="column_maker_0030",
                description=column_maker_repository_description,
                long_description=column_maker_repository_long_description,
                owner=common.test_user_1_name,
                category=category_0030,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                column_maker_repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0030_create_emboss_5_repository(self):
        """Create and populate the emboss_5_0030 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0030 = self.create_category(
                name="Test 0030 Repository Dependency Revisions",
                description="Testing repository dependencies by revision.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            emboss_5_repository = self.get_or_create_repository(
                name="emboss_5_0030",
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category_0030,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                emboss_5_repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss.tar",
            )

    def test_0035_create_emboss_6_repository(self):
        """Create and populate the emboss_6_0030 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0030 = self.create_category(
                name="Test 0030 Repository Dependency Revisions",
                description="Testing repository dependencies by revision.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            emboss_6_repository = self.get_or_create_repository(
                name="emboss_6_0030",
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category_0030,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                emboss_6_repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss.tar",
            )

    def test_0040_create_emboss_0030_repository(self):
        """Create and populate the emboss_0030 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0030 = self.create_category(
                name="Test 0030 Repository Dependency Revisions",
                description="Testing repository dependencies by revision.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            emboss_repository = self.get_or_create_repository(
                name="emboss_0030",
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category_0030,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                emboss_repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss.tar",
            )

    def test_0045_create_repository_dependencies_for_0030(self):
        """Create the dependency structure for test 0030."""
        if running_standalone:
            column_maker_repository = self._get_repository_by_name_and_owner(
                "column_maker_0030", common.test_user_1_name
            )
            emboss_repository = self._get_repository_by_name_and_owner("emboss_0030", common.test_user_1_name)
            emboss_5_repository = self._get_repository_by_name_and_owner("emboss_5_0030", common.test_user_1_name)
            emboss_6_repository = self._get_repository_by_name_and_owner("emboss_6_0030", common.test_user_1_name)
            repository_dependencies_path = self.generate_temp_path("test_0330", additional_paths=["emboss"])
            column_maker_tuple = (
                self.url,
                column_maker_repository.name,
                column_maker_repository.owner,
                self.get_repository_tip(column_maker_repository),
            )
            emboss_5_tuple = (
                self.url,
                emboss_5_repository.name,
                emboss_5_repository.owner,
                self.get_repository_tip(emboss_5_repository),
            )
            emboss_6_tuple = (
                self.url,
                emboss_6_repository.name,
                emboss_6_repository.owner,
                self.get_repository_tip(emboss_6_repository),
            )
            self.create_repository_dependency(
                repository=emboss_5_repository,
                repository_tuples=[column_maker_tuple],
                filepath=repository_dependencies_path,
            )
            self.create_repository_dependency(
                repository=emboss_6_repository,
                repository_tuples=[column_maker_tuple],
                filepath=repository_dependencies_path,
            )
            self.create_repository_dependency(
                repository=emboss_repository, repository_tuples=[emboss_5_tuple], filepath=repository_dependencies_path
            )
            self.create_repository_dependency(
                repository=emboss_repository, repository_tuples=[emboss_6_tuple], filepath=repository_dependencies_path
            )

    def test_0050_create_freebayes_repository(self):
        """Create and populate the freebayes_0040 repository."""
        self.login(email=common.admin_email, username=common.admin_username)
        category_0040 = self.create_category(
            name="test_0040_repository_circular_dependencies",
            description="Testing handling of circular repository dependencies.",
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name="freebayes_0040",
            description="Galaxy's freebayes tool",
            long_description="Long description of Galaxy's freebayes tool",
            owner=common.test_user_1_name,
            category=category_0040,
            strings_displayed=[],
        )
        if running_standalone:
            self.commit_tar_to_repository(
                repository,
                "freebayes/freebayes.tar",
                commit_message="Uploaded freebayes tarball.",
            )

    def test_0055_create_filtering_repository(self):
        """Create and populate the filtering_0040 repository."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category_0040 = self.create_category(
                name="test_0040_repository_circular_dependencies",
                description="Testing handling of circular repository dependencies.",
            )
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name="filtering_0040",
                description="Galaxy's filtering tool",
                long_description="Long description of Galaxy's filtering tool",
                owner=common.test_user_1_name,
                category=category_0040,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "filtering/filtering_1.1.0.tar",
                commit_message="Uploaded filtering 1.1.0 tarball.",
            )

    def test_0060_create_dependency_structure(self):
        """Create the dependency structure for test 0040."""
        if running_standalone:
            freebayes_repository = self._get_repository_by_name_and_owner("freebayes_0040", common.test_user_1_name)
            filtering_repository = self._get_repository_by_name_and_owner("filtering_0040", common.test_user_1_name)
            repository_dependencies_path = self.generate_temp_path("test_0340", additional_paths=["dependencies"])
            freebayes_tuple = (
                self.url,
                freebayes_repository.name,
                freebayes_repository.owner,
                self.get_repository_tip(freebayes_repository),
            )
            filtering_tuple = (
                self.url,
                filtering_repository.name,
                filtering_repository.owner,
                self.get_repository_tip(filtering_repository),
            )
            self.create_repository_dependency(
                repository=filtering_repository,
                repository_tuples=[freebayes_tuple],
                filepath=repository_dependencies_path,
            )
            self.create_repository_dependency(
                repository=freebayes_repository,
                repository_tuples=[filtering_tuple],
                filepath=repository_dependencies_path,
            )

    def test_0065_create_convert_repository(self):
        """Create and populate convert_chars_0050."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category = self.create_category(name=category_0050_name, description=category_0050_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name=convert_repository_name,
                description=convert_repository_description,
                long_description=convert_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "convert_chars/convert_chars.tar",
                commit_message="Uploaded convert_chars tarball.",
            )

    def test_0070_create_column_repository(self):
        """Create and populate convert_chars_0050."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category = self.create_category(name=category_0050_name, description=category_0050_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name=column_repository_name,
                description=column_repository_description,
                long_description=column_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0075_create_emboss_datatypes_repository(self):
        """noop now..."""

    def test_0080_create_emboss_repository(self):
        """Create and populate emboss_0050."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category = self.create_category(name=category_0050_name, description=category_0050_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name=emboss_repository_name,
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss tarball.",
            )

    def test_0085_create_filtering_repository(self):
        """Create and populate filtering_0050."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category = self.create_category(name=category_0050_name, description=category_0050_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            filtering_repository = self.get_or_create_repository(
                name=filtering_repository_name,
                description=filtering_repository_description,
                long_description=filtering_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                filtering_repository,
                "filtering/filtering_1.1.0.tar",
                commit_message="Uploaded filtering 1.1.0 tarball.",
            )

    def test_0090_create_freebayes_repository(self):
        """Create and populate freebayes_0050."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category = self.create_category(name=category_0050_name, description=category_0050_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name=freebayes_repository_name,
                description=freebayes_repository_description,
                long_description=freebayes_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "freebayes/freebayes.tar",
                commit_message="Uploaded freebayes tarball.",
            )

    def test_0095_create_bismark_repository(self):
        """Create and populate bismark_0050."""
        if running_standalone:
            self.login(email=common.admin_email, username=common.admin_username)
            category = self.create_category(name=category_0050_name, description=category_0050_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name=bismark_repository_name,
                description=bismark_repository_description,
                long_description=bismark_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.user_populator().setup_bismark_repo(repository, end=1)

    def test_0100_create_and_upload_dependency_definitions(self):
        """Create the dependency structure for test 0050."""
        if running_standalone:
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
            convert_repository = self._get_repository_by_name_and_owner(
                convert_repository_name, common.test_user_1_name
            )
            emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
            filtering_repository = self._get_repository_by_name_and_owner(
                filtering_repository_name, common.test_user_1_name
            )
            freebayes_repository = self._get_repository_by_name_and_owner(
                freebayes_repository_name, common.test_user_1_name
            )
            bismark_repository = self._get_repository_by_name_and_owner(
                bismark_repository_name, common.test_user_1_name
            )
            dependency_xml_path = self.generate_temp_path("test_0050", additional_paths=["freebayes"])
            # convert_chars depends on column_maker
            # column_maker depends on convert_chars
            # emboss depends on emboss_datatypes
            # emboss_datatypes depends on bismark
            # freebayes depends on freebayes, emboss, emboss_datatypes, and column_maker
            # filtering depends on emboss
            column_tuple = (
                self.url,
                column_repository.name,
                column_repository.owner,
                self.get_repository_tip(column_repository),
            )
            convert_tuple = (
                self.url,
                convert_repository.name,
                convert_repository.owner,
                self.get_repository_tip(convert_repository),
            )
            freebayes_tuple = (
                self.url,
                freebayes_repository.name,
                freebayes_repository.owner,
                self.get_repository_tip(freebayes_repository),
            )
            emboss_tuple = (
                self.url,
                emboss_repository.name,
                emboss_repository.owner,
                self.get_repository_tip(emboss_repository),
            )
            bismark_tuple = (
                self.url,
                bismark_repository.name,
                bismark_repository.owner,
                self.get_repository_tip(bismark_repository),
            )
            self.create_repository_dependency(
                repository=convert_repository, repository_tuples=[column_tuple], filepath=dependency_xml_path
            )
            self.create_repository_dependency(
                repository=column_repository, repository_tuples=[convert_tuple], filepath=dependency_xml_path
            )
            self.create_repository_dependency(
                repository=emboss_repository, repository_tuples=[bismark_tuple], filepath=dependency_xml_path
            )
            self.create_repository_dependency(
                repository=freebayes_repository,
                repository_tuples=[freebayes_tuple, bismark_tuple, emboss_tuple, column_tuple],
                filepath=dependency_xml_path,
            )
            self.create_repository_dependency(
                repository=filtering_repository, repository_tuples=[emboss_tuple], filepath=dependency_xml_path
            )

    @pytest.mark.xfail
    def test_0110_reset_metadata_on_all_repositories(self):
        """Reset metadata on all repositories, then verify that it has not changed."""
        self.login(email=common.admin_email, username=common.admin_username)
        old_metadata: dict[str, dict] = {}
        new_metadata: dict[str, dict] = {}
        repositories = self.test_db_util.get_all_repositories()
        for repository in repositories:
            old_metadata[self.security.encode_id(repository.id)] = {}
            for metadata in self.get_repository_metadata_for_db_object(repository):
                old_metadata[self.security.encode_id(repository.id)][metadata.changeset_revision] = metadata.metadata
        self.reset_metadata_on_selected_repositories(list(old_metadata.keys()))
        for repository in repositories:
            new_metadata[self.security.encode_id(repository.id)] = {}
            for metadata in self.get_repository_metadata_for_db_object(repository):
                new_metadata[self.security.encode_id(repository.id)][metadata.changeset_revision] = metadata.metadata
            if (
                old_metadata[self.security.encode_id(repository.id)]
                != new_metadata[self.security.encode_id(repository.id)]
            ):
                raise AssertionError(f"Metadata changed after reset for repository {repository.name}.")



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

repositories = dict(
    freebayes=dict(
        name="package_freebayes_0550",
        description="Description for package_freebayes_0550",
        long_description="Long description for package_freebayes_0550",
    ),
    samtools=dict(
        name="package_samtools_0550",
        description="Description for package_samtools_0550",
        long_description="Long description for package_samtools_0550",
    ),
    filtering=dict(
        name="filtering_0550",
        description="Description for filtering_0550",
        long_description="Long description for filtering_0550",
    ),
)

category_name = "Test 0550"
category_description = "Verify metadata updates"

"""
1. Create repository package_freebayes_0550.

2. Create repository package_samtools_0550.

3. Create repository filtering_0550.

4. Create dependency on package_freebayes_0550 for filtering_0550.

5. Create dependency on package_samtools_0550 for filtering_0550.

6. Update package_freebayes_0550 and package_samtools_0550.

5. Load /api/repositories/{filtering_0550}.id/metadata and verify contents.
"""


class TestGetUpdatedMetadata(ShedTwillTestCase):
    """Verify that updated repositories still have correct dependency links."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_freebayes_repository(self):
        """Create and populate package_freebayes_0550."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        # Create a repository named package_freebayes_0550 owned by user1.
        freebayes_repository = self.get_or_create_repository(
            name=repositories["freebayes"]["name"],
            description=repositories["freebayes"]["description"],
            long_description=repositories["freebayes"]["long_description"],
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            freebayes_repository,
            "0550_files/package_freebayes_1_0550.tgz",
        )
        if not self.is_v2:
            # Visit the manage repository page for package_freebayes_0_5_9_0100.
            self.display_manage_repository_page(
                freebayes_repository, strings_displayed=["Tool dependencies", "will not be", "to this repository"]
            )

    def test_0010_create_samtools_repository(self):
        """Create and populate the package_samtools_0550 repository."""
        category = self.create_category(name=category_name, description=category_description)
        samtools_repository = self.get_or_create_repository(
            name=repositories["samtools"]["name"],
            description=repositories["samtools"]["description"],
            long_description=repositories["samtools"]["long_description"],
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            samtools_repository,
            "0550_files/package_samtools_1_0550.tgz",
            commit_message="Uploaded samtools 1.0.",
        )

    def test_0015_create_filtering_repository(self):
        """Create the filtering_0550 repository."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=repositories["filtering"]["name"],
            description=repositories["filtering"]["description"],
            long_description=repositories["filtering"]["long_description"],
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "0550_files/filtering_1.0.tgz",
            commit_message="Uploaded filtering 1.0.",
        )

    def test_0020_check_repository_dependency(self):
        """Make filtering depend on samtools and freebayes."""
        freebayes = self._get_repository_by_name_and_owner(repositories["freebayes"]["name"], common.test_user_1_name)
        samtools = self._get_repository_by_name_and_owner(repositories["samtools"]["name"], common.test_user_1_name)
        filtering = self._get_repository_by_name_and_owner(repositories["filtering"]["name"], common.test_user_1_name)
        strings_displayed = [freebayes.id, samtools.id]
        if not self.is_v2:
            self.display_manage_repository_page(filtering, strings_displayed=strings_displayed)

    def test_0025_update_dependent_repositories(self):
        """
        Update freebayes and samtools, load the API endpoint again.
        """
        freebayes = self._get_repository_by_name_and_owner(repositories["freebayes"]["name"], common.test_user_1_name)
        samtools = self._get_repository_by_name_and_owner(repositories["samtools"]["name"], common.test_user_1_name)
        filtering = self._get_repository_by_name_and_owner(repositories["filtering"]["name"], common.test_user_1_name)
        self.commit_tar_to_repository(
            freebayes,
            "0550_files/package_freebayes_2_0550.tgz",
            commit_message="Uploaded freebayes 2.0.",
        )
        self.commit_tar_to_repository(
            samtools,
            "0550_files/package_samtools_2_0550.tgz",
            commit_message="Uploaded samtools 2.0.",
        )
        strings_displayed = [
            repositories["freebayes"]["name"],
            repositories["samtools"]["name"],
            repositories["filtering"]["name"],
        ]
        self.fetch_repository_metadata(filtering, strings_displayed=strings_displayed, strings_not_displayed=None)



import logging
import os
import shutil
import tempfile
from typing import (
    cast,
    IO,
    Optional,
    Union,
)

from fastapi import (
    Body,
    Depends,
    Request,
    Response,
    status,
    UploadFile,
)
from starlette.datastructures import UploadFile as StarletteUploadFile

from galaxy.exceptions import (
    ActionInputError,
    InsufficientPermissionsException,
    RequestParameterInvalidException,
)
from galaxy.webapps.galaxy.api import as_form
from tool_shed.context import SessionRequestContext
from tool_shed.managers.repositories import (
    can_manage_repo,
    can_update_repo,
    check_updates,
    create_repository,
    ensure_can_manage,
    get_install_info,
    get_ordered_installable_revisions,
    get_repository_metadata_dict,
    get_repository_metadata_for_management,
    index_repositories,
    index_repositories_paginated,
    IndexRequest,
    PaginatedIndexRequest,
    readmes,
    reset_metadata_on_repositories,
    reset_metadata_on_repository,
    search,
    to_detailed_model,
    to_model,
    UpdatesRequest,
    upload_tar_and_set_metadata,
)
from tool_shed.structured_app import ToolShedApp
from tool_shed.util.repository_util import (
    get_repository_in_tool_shed,
    update_validated_repository,
)
from tool_shed_client.schema import (
    CreateRepositoryRequest,
    DetailedRepository,
    from_legacy_install_info,
    IndexSortByType,
    InstallInfo,
    PaginatedRepositoryIndexResults,
    Repository,
    RepositoryMetadata,
    RepositoryPermissions,
    RepositoryRevisionReadmes,
    RepositorySearchResults,
    RepositoryUpdate,
    RepositoryUpdateRequest,
    ResetMetadataOnRepositoriesRequest,
    ResetMetadataOnRepositoriesResponse,
    ResetMetadataOnRepositoryRequest,
    ResetMetadataOnRepositoryResponse,
    UpdateRepositoryRequest,
    ValidRepostiroyUpdateMessage,
)
from . import (
    ChangesetRevisionPathParam,
    CommitMessageQueryParam,
    depend_on_either_json_or_form_data,
    depends,
    DependsOnTrans,
    DownloadableQueryParam,
    OptionalHexlifyParam,
    OptionalRepositoryIdParam,
    OptionalRepositoryNameParam,
    OptionalRepositoryOwnerParam,
    RepositoryIdPathParam,
    RepositoryIndexCategoryQueryParam,
    RepositoryIndexDeletedQueryParam,
    RepositoryIndexFilterParam,
    RepositoryIndexNameQueryParam,
    RepositoryIndexOwnerQueryParam,
    RepositoryIndexQueryParam,
    RepositoryIndexSortByParam,
    RepositoryIndexSortDescParam,
    RepositorySearchPageQueryParam,
    RepositorySearchPageSizeQueryParam,
    RequiredChangesetParam,
    RequiredRepoNameParam,
    RequiredRepoOwnerParam,
    RequiredRepositoryChangesetRevisionParam,
    Router,
    UsernameIdPathParam,
)

log = logging.getLogger(__name__)


router = Router(tags=["repositories"])

IndexResponse = Union[RepositorySearchResults, list[Repository], PaginatedRepositoryIndexResults]


@as_form
class RepositoryUpdateRequestFormData(RepositoryUpdateRequest):
    pass


@router.cbv
class FastAPIRepositories:
    app: ToolShedApp = depends(ToolShedApp)

    @router.get(
        "/api/repositories",
        description="Get a list of repositories or perform a search.",
        operation_id="repositories__index",
    )
    def index(
        self,
        q: Optional[str] = RepositoryIndexQueryParam,
        filter: Optional[str] = RepositoryIndexFilterParam,
        page: Optional[int] = RepositorySearchPageQueryParam,
        page_size: Optional[int] = RepositorySearchPageSizeQueryParam,
        deleted: Optional[bool] = RepositoryIndexDeletedQueryParam,
        owner: Optional[str] = RepositoryIndexOwnerQueryParam,
        name: Optional[str] = RepositoryIndexNameQueryParam,
        category_id: Optional[str] = RepositoryIndexCategoryQueryParam,
        sort_desc: Optional[bool] = RepositoryIndexSortDescParam,
        sort_by: Optional[IndexSortByType] = RepositoryIndexSortByParam,
        trans: SessionRequestContext = DependsOnTrans,
    ) -> IndexResponse:

        if q and filter:
            raise RequestParameterInvalidException(
                "Cannot specify both the 'q' and 'filter' parameter at the same time."
            )

        if q:
            page = page or 1
            assert page_size is not None
            search_results = search(trans, q, page, page_size)
            return RepositorySearchResults(**search_results)
        # See API notes - was added in https://github.com/galaxyproject/galaxy/pull/3626/files
        # but I think is currently unused. So probably we should just drop it until someone
        # complains.
        # elif params.tool_ids:
        #    response = index_tool_ids(self.app, params.tool_ids)
        #    return response
        elif page:
            assert page_size is not None
            paginated_index_request = PaginatedIndexRequest(
                page=page,
                page_size=page_size,
                owner=owner,
                name=name,
                deleted=deleted or False,
                filter=filter,
                category_id=category_id,
                sort_by=sort_by,
                sort_desc=sort_desc,
            )
            paginated_repositories = index_repositories_paginated(self.app, paginated_index_request)
            return paginated_repositories
        else:
            index_request = IndexRequest(
                owner=owner,
                name=name,
                deleted=deleted or False,
                filter=filter,
                category_id=category_id,
                sort_by=sort_by,
                sort_desc=sort_desc,
            )
            repositories = index_repositories(self.app, index_request)
            return [to_model(self.app, r) for r in repositories]

    @router.get(
        "/api/repositories/get_repository_revision_install_info",
        description="Get information used by the install client to install this repository.",
        operation_id="repositories__legacy_install_info",
    )
    def legacy_install_info(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        name: str = RequiredRepoNameParam,
        owner: str = RequiredRepoOwnerParam,
        changeset_revision: str = RequiredChangesetParam,
    ) -> list:
        legacy_install_info = get_install_info(
            trans,
            name,
            owner,
            changeset_revision,
        )
        return list(legacy_install_info)

    @router.get(
        "/api/repositories/install_info",
        description="Get information used by the install client to install this repository.",
        operation_id="repositories__install_info",
    )
    def install_info(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        name: str = RequiredRepoNameParam,
        owner: str = RequiredRepoOwnerParam,
        changeset_revision: str = RequiredChangesetParam,
    ) -> InstallInfo:
        # A less problematic version of the above API, but I guess we
        # need to maintain the older version for older Galaxy API clients
        # for... sometime... or forever.
        legacy_install_info = get_install_info(
            trans,
            name,
            owner,
            changeset_revision,
        )
        return from_legacy_install_info(legacy_install_info)

    @router.get(
        "/api/repositories/{encoded_repository_id}/metadata",
        description="Get information about repository metadata",
        operation_id="repositories__metadata",
        # See comment below.
        # response_model=RepositoryMetadata,
    )
    def metadata(
        self,
        encoded_repository_id: str = RepositoryIdPathParam,
        downloadable_only: bool = DownloadableQueryParam,
    ) -> dict:
        recursive = True
        as_dict = get_repository_metadata_dict(self.app, encoded_repository_id, recursive, downloadable_only)
        # fails 1020 if we try to use the model - I guess repository dependencies
        # are getting lost
        return as_dict

    @router.get(
        "/api_internal/repositories/{encoded_repository_id}/metadata",
        description="Get information about repository metadata",
        operation_id="repositories__internal_metadata",
        response_model=RepositoryMetadata,
    )
    def metadata_internal(
        self,
        encoded_repository_id: str = RepositoryIdPathParam,
        downloadable_only: bool = DownloadableQueryParam,
    ) -> RepositoryMetadata:
        recursive = True
        as_dict = get_repository_metadata_dict(self.app, encoded_repository_id, recursive, downloadable_only)
        return RepositoryMetadata(root=as_dict)

    @router.get(
        "/api/repositories/get_ordered_installable_revisions",
        description="Get an ordered list of the repository changeset revisions that are installable",
        operation_id="repositories__get_ordered_installable_revisions",
    )
    def get_ordered_installable_revisions(
        self,
        owner: Optional[str] = OptionalRepositoryOwnerParam,
        name: Optional[str] = OptionalRepositoryNameParam,
        tsr_id: Optional[str] = OptionalRepositoryIdParam,
    ) -> list[str]:
        return get_ordered_installable_revisions(self.app, name, owner, tsr_id)

    @router.post(
        "/api/repositories/reset_metadata_on_repository",
        description="reset metadata on a repository",
        operation_id="repositories__reset_legacy",
    )
    def reset_metadata_on_repository_legacy(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        request: ResetMetadataOnRepositoryRequest = depend_on_either_json_or_form_data(
            ResetMetadataOnRepositoryRequest
        ),
    ) -> ResetMetadataOnRepositoryResponse:
        return reset_metadata_on_repository(trans, request.repository_id)

    @router.post(
        "/api/repositories/{encoded_repository_id}/reset_metadata",
        description="reset metadata on a repository",
        operation_id="repositories__reset",
    )
    def reset_metadata_on_repository(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
    ) -> ResetMetadataOnRepositoryResponse:
        return reset_metadata_on_repository(trans, encoded_repository_id)

    @router.post(
        "/api/repositories/reset_metadata_on_repositories",
        description="reset metadata on all of your repositories",
        operation_id="repositories__reset_all",
    )
    def reset_metadata_on_repositories(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        request: ResetMetadataOnRepositoriesRequest = depend_on_either_json_or_form_data(
            ResetMetadataOnRepositoriesRequest
        ),
    ) -> ResetMetadataOnRepositoriesResponse:
        return reset_metadata_on_repositories(trans, request)

    @router.get(
        "/api/repositories/updates",
        operation_id="repositories__update",
    )
    @router.get(
        "/api/repositories/updates/",
    )
    def updates(
        self,
        owner: Optional[str] = OptionalRepositoryOwnerParam,
        name: Optional[str] = OptionalRepositoryNameParam,
        changeset_revision: str = RequiredRepositoryChangesetRevisionParam,
        hexlify: Optional[bool] = OptionalHexlifyParam,
    ):
        request = UpdatesRequest(
            name=name,
            owner=owner,
            changeset_revision=changeset_revision,
            hexlify=hexlify,
        )
        return Response(content=check_updates(self.app, request))

    @router.post(
        "/api/repositories",
        description="create a new repository",
        operation_id="repositories__create",
    )
    def create(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        request: CreateRepositoryRequest = Body(...),
    ) -> Repository:
        db_repository = create_repository(
            trans,
            request,
        )
        return to_model(self.app, db_repository)

    @router.get(
        "/api/repositories/{encoded_repository_id}",
        operation_id="repositories__show",
    )
    def show(
        self,
        encoded_repository_id: str = RepositoryIdPathParam,
    ) -> DetailedRepository:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        return to_detailed_model(self.app, repository)

    @router.put(
        "/api/repositories/{encoded_repository_id}",
        operation_id="repositories__update_repository",
    )
    def update_repository(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
        request: UpdateRepositoryRequest = Body(...),
    ) -> DetailedRepository:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        ensure_can_manage(trans, repository)

        # may want to set some of these to null, so we're using the exclude_unset feature
        # to just serialize the ones we want to use to a dictionary.
        update_dictionary = request.model_dump(exclude_unset=True)
        repo_result, message = update_validated_repository(trans, repository, **update_dictionary)
        if repo_result is None:
            raise ActionInputError(message)

        return to_detailed_model(self.app, repository)

    @router.get(
        "/api/repositories/{encoded_repository_id}/permissions",
        operation_id="repositories__permissions",
    )
    def permissions(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
    ) -> RepositoryPermissions:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        if not can_update_repo(trans, repository):
            raise InsufficientPermissionsException(
                "You do not have permission to inspect repository repository permissions."
            )
        return RepositoryPermissions(
            allow_push=trans.app.security_agent.usernames_that_can_push(repository),
            can_manage=can_manage_repo(trans, repository),
            can_push=can_update_repo(trans, repository),
        )

    @router.get(
        "/api/repositories/{encoded_repository_id}/allow_push",
        operation_id="repositories__show_allow_push",
    )
    def show_allow_push(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
    ) -> list[str]:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        ensure_can_manage(trans, repository)
        return trans.app.security_agent.usernames_that_can_push(repository)

    @router.post(
        "/api/repositories/{encoded_repository_id}/allow_push/{username}",
        operation_id="repositories__add_allow_push",
    )
    def add_allow_push(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
        username: str = UsernameIdPathParam,
    ) -> list[str]:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        if not can_manage_repo(trans, repository):
            raise InsufficientPermissionsException("You do not have permission to update this repository.")
        repository.set_allow_push([username])
        return trans.app.security_agent.usernames_that_can_push(repository)

    @router.put(
        "/api/repositories/{encoded_repository_id}/revisions/{changeset_revision}/malicious",
        operation_id="repositories__set_malicious",
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def set_malicious(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
        changeset_revision: str = ChangesetRevisionPathParam,
    ):
        repository_metadata = get_repository_metadata_for_management(trans, encoded_repository_id, changeset_revision)
        repository_metadata.malicious = True
        trans.sa_session.add(repository_metadata)
        trans.sa_session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.delete(
        "/api/repositories/{encoded_repository_id}/revisions/{changeset_revision}/malicious",
        operation_id="repositories__unset_malicious",
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def unset_malicious(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
        changeset_revision: str = ChangesetRevisionPathParam,
    ):
        repository_metadata = get_repository_metadata_for_management(trans, encoded_repository_id, changeset_revision)
        repository_metadata.malicious = False
        trans.sa_session.add(repository_metadata)
        trans.sa_session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.put(
        "/api/repositories/{encoded_repository_id}/deprecated",
        operation_id="repositories__set_deprecated",
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def set_deprecated(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
    ):
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        ensure_can_manage(trans, repository)
        repository.deprecated = True
        trans.sa_session.add(repository)
        trans.sa_session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.delete(
        "/api/repositories/{encoded_repository_id}/deprecated",
        operation_id="repositories__unset_deprecated",
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def unset_deprecated(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
    ):
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        ensure_can_manage(trans, repository)
        repository.deprecated = False
        trans.sa_session.add(repository)
        trans.sa_session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.delete(
        "/api/repositories/{encoded_repository_id}/allow_push/{username}",
        operation_id="repositories__remove_allow_push",
    )
    def remove_allow_push(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_repository_id: str = RepositoryIdPathParam,
        username: str = UsernameIdPathParam,
    ) -> list[str]:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        if not can_manage_repo(trans, repository):
            raise InsufficientPermissionsException("You do not have permission to update this repository.")
        repository.set_allow_push(None, remove_auth=username)
        return trans.app.security_agent.usernames_that_can_push(repository)

    @router.post(
        "/api/repositories/{encoded_repository_id}/changeset_revision",
        description="upload new revision to the repository",
        operation_id="repositories__create_revision",
    )
    async def create_changeset_revision(
        self,
        request: Request,
        encoded_repository_id: str = RepositoryIdPathParam,
        commit_message: Optional[str] = CommitMessageQueryParam,
        trans: SessionRequestContext = DependsOnTrans,
        files: Optional[list[UploadFile]] = None,
        revision_request: RepositoryUpdateRequest = Depends(RepositoryUpdateRequestFormData.as_form),  # type: ignore[attr-defined]
    ) -> RepositoryUpdate:
        try:
            # Code stolen from Marius' work in Galaxy's Tools API.

            files2: list[StarletteUploadFile] = cast(list[StarletteUploadFile], files or [])
            # FastAPI's UploadFile is a very light wrapper around starlette's UploadFile
            if not files2:
                data = await request.form()
                for value in data.values():
                    if isinstance(value, StarletteUploadFile):
                        files2.append(value)

            repository = get_repository_in_tool_shed(self.app, encoded_repository_id)

            if not can_update_repo(trans, repository):
                raise InsufficientPermissionsException("You do not have permission to update this repository.")

            assert trans.user
            assert files2
            the_file = files2[0]
            with tempfile.NamedTemporaryFile(
                dir=trans.app.config.new_file_path, prefix="upload_file_data_", delete=False
            ) as dest:
                upload_file_like: IO[bytes] = the_file.file
                shutil.copyfileobj(upload_file_like, dest)
            the_file.file.close()
            filename = dest.name
            try:
                message = upload_tar_and_set_metadata(
                    trans,
                    trans.request.host,
                    repository,
                    filename,
                    commit_message or revision_request.commit_message or "Uploaded",
                )
                return RepositoryUpdate(root=ValidRepostiroyUpdateMessage(message=message))
            finally:
                if os.path.exists(filename):
                    os.remove(filename)
        except Exception:
            log.exception("Problem in here...")
            raise

    @router.get(
        "/api/repositories/{encoded_repository_id}/revisions/{changeset_revision}/readmes",
        description="fetch readmes for repository revision",
        operation_id="repositories__readmes",
        response_model=RepositoryRevisionReadmes,
    )
    def get_readmes(
        self,
        encoded_repository_id: str = RepositoryIdPathParam,
        changeset_revision: str = ChangesetRevisionPathParam,
    ) -> dict:
        repository = get_repository_in_tool_shed(self.app, encoded_repository_id)
        return readmes(self.app, repository, changeset_revision)



import logging
import os
from pathlib import Path
from typing import (
    Any,
    cast,
    Optional,
)

from a2wsgi import WSGIMiddleware
from fastapi import (
    Depends,
    FastAPI,
)
from fastapi.responses import (
    HTMLResponse,
    RedirectResponse,
)
from fastapi.staticfiles import StaticFiles
from slowapi import (
    _rate_limit_exceeded_handler,
    Limiter,
)
from slowapi.errors import RateLimitExceeded
from slowapi.util import get_remote_address

from galaxy.webapps.base.api import (
    add_exception_handler,
    add_request_id_middleware,
    include_all_package_routers,
)
from galaxy.webapps.openapi.utils import get_openapi
from tool_shed.webapp.api2 import (
    ensure_valid_session,
    get_trans,
)

log = logging.getLogger(__name__)

api_tags_metadata = [
    {
        "name": "authenticate",
        "description": "Authentication-related endpoints.",
    },
    {
        "name": "categories",
        "description": "Category-related endpoints.",
    },
    {
        "name": "repositories",
        "description": "Repository-related endpoints.",
    },
    {
        "name": "users",
        "description": "User-related endpoints.",
    },
    {"name": "undocumented", "description": "API routes that have not yet been ported to FastAPI."},
]

# Set this if asset handling should be sent to vite.
# Run vite with:
#   yarn dev
# Start tool shed with:
#   TOOL_SHED_VITE_PORT=4040 TOOL_SHED_API_VERSION=v2 ./run_tool_shed.sh
TOOL_SHED_VITE_PORT: Optional[str] = os.environ.get("TOOL_SHED_VITE_PORT", None)
TOOL_SHED_FRONTEND_TARGET: str = os.environ.get("TOOL_SHED_FRONTEND_TARGET") or "auto"  # auto, src, or node
TOOL_SHED_USE_HMR: bool = TOOL_SHED_VITE_PORT is not None
WEBAPP_DIR = Path(__file__).parent.resolve()
FRONTEND = WEBAPP_DIR / "frontend"
FRONTEND_DIST = FRONTEND / "dist"
INSTALLED_FRONTEND = WEBAPP_DIR / "node_modules" / "@galaxyproject" / "tool-shed-frontend" / "dist"
INDEX_FILENAME = "index.html"


def find_frontend_target() -> Path:
    src_target = FRONTEND_DIST
    node_target = INSTALLED_FRONTEND
    if TOOL_SHED_FRONTEND_TARGET == "src":
        return src_target
    elif TOOL_SHED_FRONTEND_TARGET == "node":
        return node_target
    elif src_target.exists():
        return src_target
    else:
        return node_target


def frontend_controller(app):
    shed_entry_point = "main.ts"
    vite_runtime = "@vite/client"

    def index(trans=Depends(get_trans)):
        if TOOL_SHED_USE_HMR:
            if TOOL_SHED_FRONTEND_TARGET != "auto":
                raise Exception("Cannot configure HMR and with this frontend target.")
            index = FRONTEND / INDEX_FILENAME
            index_html = index.read_text()
            index_html = index_html.replace(
                f"""<script type="module" src="/src/{shed_entry_point}"></script>""",
                f"""<script type="module" src="http://localhost:{TOOL_SHED_VITE_PORT}/{vite_runtime}"></script><script type="module" src="http://localhost:{TOOL_SHED_VITE_PORT}/src/{shed_entry_point}"></script>""",
            )
        else:
            index = find_frontend_target() / INDEX_FILENAME
            index_html = index.read_text()
        ensure_valid_session(trans)
        cookie = trans.session_csrf_token
        r: HTMLResponse = cast(HTMLResponse, trans.response)
        r.set_cookie("session_csrf_token", cookie)
        return index_html

    return app, index


def redirect_route(app, from_url: str, to_url: str):
    @app.get(from_url)
    def redirect():
        return RedirectResponse(to_url)


def frontend_route(controller, path):
    app, index = controller
    app.get(path, response_class=HTMLResponse)(index)


FRONT_END_ROUTES = [
    "/",
    "/admin",
    "/login",
    "/register",
    "/logout_success",
    "/login_success",
    "/registration_success",
    "/help",
    "/repositories_by_search",
    "/repositories_by_category",
    "/repositories_by_category/{category_id}",
    "/repositories_by_owner",
    "/repositories_by_owner/{username}",
    "/repositories/{repository_id}",
    "/repositories_search",
    "/tools/{trs_tool_id}/versions/{version}",
    "/_component_showcase",
    "/user/api_key",
    "/user/change_password",
    "/user/change_password_success",
    "/view/{username}",
    "/view/{username}/{repository_name}",
    "/view/{username}/{repository_name}/{changeset_revision}",
]
LEGACY_ROUTES = {
    "/user/create": "/register",  # for twilltestcase
    "/user/login": "/login",  # for twilltestcase
}


limiter = Limiter(key_func=get_remote_address)


def initialize_fast_app(gx_webapp, tool_shed_app):
    app = get_fastapi_instance()
    add_exception_handler(app)
    add_request_id_middleware(app)
    app.state.limiter = limiter
    app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)  # type: ignore[arg-type]
    from .config import SHED_API_VERSION

    def mount_static(directory: Path):
        name = directory.name
        if directory.exists():
            app.mount(f"/{name}", StaticFiles(directory=directory), name=name)

    if SHED_API_VERSION == "v2":
        controller = frontend_controller(app)
        for route in FRONT_END_ROUTES:
            frontend_route(controller, route)

        for from_route, to_route in LEGACY_ROUTES.items():
            redirect_route(app, from_route, to_route)

        mount_static(FRONTEND / "static")
        if TOOL_SHED_USE_HMR:
            mount_static(FRONTEND / "node_modules")
        else:
            mount_static(find_frontend_target() / "assets")

    routes_package = "tool_shed.webapp.api" if SHED_API_VERSION == "v1" else "tool_shed.webapp.api2"
    include_all_package_routers(app, routes_package)
    wsgi_handler = WSGIMiddleware(gx_webapp)
    tool_shed_app.haltables.append(("WSGI Middleware threadpool", wsgi_handler.executor.shutdown))
    # https://github.com/abersheeran/a2wsgi/issues/44
    app.mount("/", wsgi_handler)  # type: ignore[arg-type]
    return app


def get_fastapi_instance() -> FastAPI:
    return FastAPI(
        title="Galaxy Tool Shed API",
        description=("This API allows you to manage the Tool Shed repositories."),
        docs_url="/api/docs",
        redoc_url="/api/redoc",
        tags=api_tags_metadata,
        license_info={"name": "MIT", "url": "https://github.com/galaxyproject/galaxy/blob/dev/LICENSE.txt"},
    )


def get_openapi_schema() -> dict[str, Any]:
    """
    Dumps openAPI schema without starting a full app and webserver.
    """
    app = get_fastapi_instance()
    include_all_package_routers(app, "tool_shed.webapp.api2")
    return get_openapi(
        title=app.title,
        version=app.version,
        openapi_version="3.1.0",
        description=app.description,
        routes=app.routes,
        license_info=app.license_info,
    )


__all__ = (
    "add_request_id_middleware",
    "get_openapi_schema",
    "initialize_fast_app",
)



import logging
import os

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

repository_name = "filtering_0310"
repository_description = "Galaxy's filtering tool for test 0310"
repository_long_description = "Long description of Galaxy's filtering tool for test 0310"

category_name = "Test 0310 - HTTP Repo features"
category_description = "Test 0310 for verifying the tool shed http interface to mercurial."

"""
1. Create a repository.
2. Clone the repository to a local path.
"""


class TestHgWebFeatures(ShedTwillTestCase):
    """Test http mercurial interface."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_filtering_repository(self):
        """Create and populate the filtering_0310 repository.

        We are at step 1 - Create a repository.
        Create and populate the filtering_0310 repository.
        """
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "filtering/filtering_1.1.0.tar",
            commit_message="Uploaded filtering 1.1.0.",
        )

    def test_0010_clone(self):
        """Clone the repository to a local path.

        We are at step 2 - Clone the repository to a local path.
        The repository should have the following files:

        filtering.py
        filtering.xml
        test-data/
        test-data/1.bed
        test-data/7.bed
        test-data/filter1_in3.sam
        test-data/filter1_inbad.bed
        test-data/filter1_test1.bed
        test-data/filter1_test2.bed
        test-data/filter1_test3.sam
        test-data/filter1_test4.bed
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        clone_path = self.generate_temp_path("test_0310", additional_paths=["filtering_0310", "user2"])
        self.clone_repository(repository, clone_path)
        files_in_repository = os.listdir(clone_path)
        assert "filtering.py" in files_in_repository, "File not found in repository: filtering.py"



from tool_shed_client.schema.trs import (
    Tool,
    ToolClass,
)
from tool_shed_client.trs_util import encode_identifier
from ..base.api import (
    ShedApiTestCase,
    skip_if_api_v1,
)


class TestShedToolsApi(ShedApiTestCase):
    def test_tool_search(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="toolsearch")
        populator.reindex()
        response = populator.tool_search_query("Compute")
        hit_count = len(response.hits)
        assert hit_count >= 1

        if hit_count == 1:
            # if running the test standalone, we know this repository was in the results
            # but if this tool has been installed a bunch by other tests - it might not be.
            tool_search_hit = response.find_search_hit(repository)
            assert tool_search_hit
            assert tool_search_hit.tool.id == "Add_a_column1"
            assert tool_search_hit.tool.name == "Compute"

        # ensure re-index doesn't modify number of hits (regression of an issue pre-Fall 2022)
        populator.reindex()
        populator.reindex()

        response = populator.tool_search_query("Compute")
        new_hit_count = len(response.hits)

        assert hit_count == new_hit_count

        if hit_count == 1:
            # if running the test standalone, we know this repository was in the results
            # but if this tool has been installed a bunch by other tests - it might not be.
            tool_search_hit = response.find_search_hit(repository)
            assert tool_search_hit

    @skip_if_api_v1
    def test_trs_service_info(self):
        service_info = self.api_interactor.get("ga4gh/trs/v2/service-info")
        service_info.raise_for_status()

    @skip_if_api_v1
    def test_trs_tool_classes(self):
        classes_response = self.api_interactor.get("ga4gh/trs/v2/toolClasses")
        classes_response.raise_for_status()
        classes = classes_response.json()
        assert isinstance(classes, list)
        assert len(classes) == 1
        class0 = classes[0]
        assert ToolClass(**class0)

    @skip_if_api_v1
    def test_trs_tool_list(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="toolstrsindex")
        tool_id = populator.tool_guid(self, repository, "Add_a_column1")
        tool_shed_base, encoded_tool_id = encode_identifier(tool_id)
        url = f"ga4gh/trs/v2/tools/{encoded_tool_id}"
        tool_response = self.api_interactor.get(url)
        tool_response.raise_for_status()
        assert Tool(**tool_response.json())

    @skip_if_api_v1
    def test_trs_tool_parameter_json_schema(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="toolsparameterschema")
        tool_id = populator.tool_guid(self, repository, "Add_a_column1")
        tool_shed_base, encoded_tool_id = encode_identifier(tool_id)
        url = f"tools/{encoded_tool_id}/versions/1.1.0/parameter_request_schema"
        tool_response = self.api_interactor.get(url)
        tool_response.raise_for_status()



"""
Provides factory methods to assemble the Galaxy web application
"""

import atexit
import logging
import os
from inspect import isclass
from urllib.parse import parse_qs

import routes
from paste import httpexceptions
from routes.middleware import RoutesMiddleware

import galaxy.webapps.base.webapp
from galaxy import util
from galaxy.structured_app import BasicSharedApp
from galaxy.util import asbool
from galaxy.util.properties import load_app_properties
from galaxy.web import url_for
from galaxy.web.framework.middleware.error import ErrorMiddleware
from galaxy.web.framework.middleware.request_id import RequestIDMiddleware
from galaxy.web.framework.middleware.xforwardedhost import XForwardedHostMiddleware
from galaxy.webapps.base.webapp import (
    build_url_map,
    GalaxyWebTransaction,
)
from galaxy.webapps.util import wrap_if_allowed
from .config import SHED_API_VERSION

log = logging.getLogger(__name__)


class ToolShedGalaxyWebTransaction(GalaxyWebTransaction):
    @property
    def repositories_hostname(self) -> str:
        return url_for("/", qualified=True).rstrip("/")

    def get_or_create_default_history(self):
        # tool shed has no concept of histories
        raise NotImplementedError


class CommunityWebApplication(galaxy.webapps.base.webapp.WebApplication):
    injection_aware: bool = True

    def transaction_chooser(self, environ, galaxy_app: BasicSharedApp, session_cookie: str):
        return ToolShedGalaxyWebTransaction(environ, galaxy_app, self, session_cookie)


def add_ui_controllers(webapp, app):
    """
    Search for controllers in the 'galaxy.webapps.controllers' module and add
    them to the webapp.
    """
    import tool_shed.webapp.controllers
    from galaxy.webapps.base.controller import BaseUIController

    controller_dir = tool_shed.webapp.controllers.__path__[0]
    for fname in os.listdir(controller_dir):
        if not fname.startswith("_") and fname.endswith(".py"):
            name = fname[:-3]
            module_name = f"tool_shed.webapp.controllers.{name}"
            module = __import__(module_name)
            for comp in module_name.split(".")[1:]:
                module = getattr(module, comp)
            # Look for a controller inside the modules
            for key in dir(module):
                T = getattr(module, key)
                if isclass(T) and T is not BaseUIController and issubclass(T, BaseUIController):
                    webapp.add_ui_controller(name, T(app))


def app_factory(*args, **kwargs):
    """
    Return a wsgi application serving the root object
    """
    return app_pair(*args, **kwargs)[0]


def app_pair(global_conf, load_app_kwds=None, **kwargs):
    """Return a wsgi application serving the root object"""
    # Create the Galaxy tool shed application unless passed in
    load_app_kwds = load_app_kwds or {}
    kwargs = load_app_properties(kwds=kwargs, config_prefix="TOOL_SHED_CONFIG_", **load_app_kwds)
    if "app" in kwargs:
        app = kwargs.pop("app")
        import tool_shed.webapp.app

        tool_shed.webapp.app.app = app
    else:
        try:
            import tool_shed.webapp.app

            app = tool_shed.webapp.app.UniverseApplication(global_conf=global_conf, **kwargs)
            tool_shed.webapp.app.app = app
        except Exception:
            import sys
            import traceback

            traceback.print_exc()
            sys.exit(1)
    atexit.register(app.shutdown)
    # Create the universe WSGI application
    webapp = CommunityWebApplication(app, session_cookie="galaxycommunitysession", name="tool_shed")
    add_ui_controllers(webapp, app)
    webapp.add_route("/view/{owner}", controller="repository", action="sharable_owner")
    webapp.add_route("/view/{owner}/{name}", controller="repository", action="sharable_repository")
    webapp.add_route(
        "/view/{owner}/{name}/{changeset_revision}", controller="repository", action="sharable_repository_revision"
    )
    # Handle displaying tool help images and README file images for tools contained in repositories.
    webapp.add_route(
        "/repository/static/images/{repository_id}/{image_file:.+?}",
        controller="repository",
        action="display_image_in_repository",
        repository_id=None,
        image_file=None,
    )
    webapp.add_route("/{controller}/{action}", action="index")
    webapp.add_route("/{action}", controller="repository", action="index")
    # Enable 'hg clone' functionality on repos by letting hgwebapp handle the request
    webapp.add_route("/repos/*path_info", controller="hg", action="handle_request", path_info="/")
    # Add the web API.  # A good resource for RESTful services - https://routes.readthedocs.io/en/latest/restful.html
    if SHED_API_VERSION == "v1":
        webapp.add_api_controllers("tool_shed.webapp.api", app)
        webapp.mapper.connect(
            "api_key_retrieval",
            "/api/authenticate/baseauth/",
            controller="authenticate",
            action="get_tool_shed_api_key",
            conditions=dict(method=["GET"]),
        )
        webapp.mapper.connect(
            "group", "/api/groups/", controller="groups", action="index", conditions=dict(method=["GET"])
        )
        webapp.mapper.connect(
            "group", "/api/groups/", controller="groups", action="create", conditions=dict(method=["POST"])
        )
        webapp.mapper.connect(
            "group", "/api/groups/{encoded_id}", controller="groups", action="show", conditions=dict(method=["GET"])
        )
        webapp.mapper.resource(
            "category",
            "categories",
            controller="categories",
            name_prefix="category_",
            path_prefix="/api",
            parent_resources=dict(member_name="category", collection_name="categories"),
        )
        webapp.mapper.connect(
            "repositories_in_category",
            "/api/categories/{category_id}/repositories",
            controller="categories",
            action="get_repositories",
            conditions=dict(method=["GET"]),
        )
        webapp.mapper.connect(
            "show_updates_for_repository",
            "/api/repositories/updates",
            controller="repositories",
            action="updates",
            conditions=dict(method=["GET"]),
        )
        webapp.mapper.resource(
            "repository",
            "repositories",
            controller="repositories",
            collection={
                "add_repository_registry_entry": "POST",
                "get_repository_revision_install_info": "GET",
                "get_ordered_installable_revisions": "GET",
                "get_installable_revisions": "GET",
                "remove_repository_registry_entry": "POST",
                "reset_metadata_on_repositories": "POST",
                "reset_metadata_on_repository": "POST",
            },
            name_prefix="repository_",
            path_prefix="/api",
            parent_resources=dict(member_name="repository", collection_name="repositories"),
        )
        webapp.mapper.resource(
            "repository_revision",
            "repository_revisions",
            member={"repository_dependencies": "GET", "export": "POST"},
            controller="repository_revisions",
            name_prefix="repository_revision_",
            path_prefix="/api",
            parent_resources=dict(member_name="repository_revision", collection_name="repository_revisions"),
        )
        webapp.mapper.resource(
            "user",
            "users",
            controller="users",
            name_prefix="user_",
            path_prefix="/api",
            parent_resources=dict(member_name="user", collection_name="users"),
        )
        webapp.mapper.connect(
            "update_repository",
            "/api/repositories/{id}",
            controller="repositories",
            action="update",
            conditions=dict(method=["PATCH", "PUT"]),
        )
        webapp.mapper.connect(
            "repository_create_changeset_revision",
            "/api/repositories/{id}/changeset_revision",
            controller="repositories",
            action="create_changeset_revision",
            conditions=dict(method=["POST"]),
        )
        webapp.mapper.connect(
            "repository_get_metadata",
            "/api/repositories/{id}/metadata",
            controller="repositories",
            action="metadata",
            conditions=dict(method=["GET"]),
        )
        webapp.mapper.connect(
            "repository_show_tools",
            "/api/repositories/{id}/{changeset}/show_tools",
            controller="repositories",
            action="show_tools",
            conditions=dict(method=["GET"]),
        )
        webapp.mapper.connect(
            "create_repository",
            "/api/repositories",
            controller="repositories",
            action="create",
            conditions=dict(method=["POST"]),
        )
        webapp.mapper.connect(
            "tools",
            "/api/tools/build_search_index",
            controller="tools",
            action="build_search_index",
            conditions=dict(method=["PUT"]),
        )
        webapp.mapper.connect(
            "tools", "/api/tools", controller="tools", action="index", conditions=dict(method=["GET"])
        )
        webapp.mapper.connect(
            "version", "/api/version", controller="configuration", action="version", conditions=dict(method=["GET"])
        )

    webapp.finalize_config()
    # Wrap the webapp in some useful middleware
    if kwargs.get("middleware", True):
        webapp = wrap_in_middleware(webapp, global_conf, app.application_stack, **kwargs)
    if asbool(kwargs.get("static_enabled", True)):
        webapp = wrap_if_allowed(webapp, app.application_stack, build_url_map, args=(global_conf,), kwargs=kwargs)
    return webapp, app


def wrap_in_middleware(app, global_conf, application_stack, **local_conf):
    """Based on the configuration wrap `app` in a set of common and useful middleware."""
    stack = application_stack
    # Merge the global and local configurations
    conf = global_conf.copy()
    conf.update(local_conf)
    # First put into place httpexceptions, which must be most closely
    # wrapped around the application (it can interact poorly with
    # other middleware):
    app = wrap_if_allowed(app, stack, httpexceptions.make_middleware, name="paste.httpexceptions", args=(conf,))
    # Create a separate mapper for redirects to prevent conflicts.
    redirect_mapper = routes.Mapper()
    redirect_mapper = _map_redirects(redirect_mapper)
    # Load the Routes middleware which we use for redirecting
    app = wrap_if_allowed(app, stack, RoutesMiddleware, args=(redirect_mapper,))
    # If we're using remote_user authentication, add middleware that
    # protects Galaxy from improperly configured authentication in the
    # upstream server
    if asbool(conf.get("use_remote_user", False)):
        from tool_shed.webapp.framework.middleware.remoteuser import RemoteUser

        app = wrap_if_allowed(
            app,
            stack,
            RemoteUser,
            kwargs=dict(
                maildomain=conf.get("remote_user_maildomain", None),
                display_servers=util.listify(conf.get("display_servers", "")),
                admin_users=conf.get("admin_users", "").split(","),
                remote_user_header=conf.get("remote_user_header", "HTTP_REMOTE_USER"),
                remote_user_secret_header=conf.get("remote_user_secret", None),
                normalize_remote_user_email=conf.get("normalize_remote_user_email", False),
            ),
        )

    # Transaction logging (apache access.log style)
    if asbool(conf.get("use_translogger", True)):
        from paste.translogger import TransLogger

        app = wrap_if_allowed(app, stack, TransLogger)

    # X-Forwarded-Host handling
    app = wrap_if_allowed(app, stack, XForwardedHostMiddleware)
    # Request ID handling
    app = wrap_if_allowed(app, stack, RequestIDMiddleware)
    # Error middleware
    app = wrap_if_allowed(app, stack, ErrorMiddleware, args=(conf,))
    return app


def _map_redirects(mapper):
    """
    Add redirect to the Routes mapper and forward the received query string.
    Subsequently when the redirect is triggered in Routes middleware the request
    will not even reach the webapp.
    """

    def forward_qs(environ, result):
        qs_dict = parse_qs(environ["QUERY_STRING"])
        for qs in qs_dict:
            result[qs] = qs_dict[qs]
        return True

    mapper.redirect(
        "/repository/status_for_installed_repository",
        "/api/repositories/updates/",
        _redirect_code="301 Moved Permanently",
        conditions=dict(function=forward_qs),
    )
    return mapper



import tarfile
from collections.abc import Iterator
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import (
    Optional,
    Union,
)

import requests
from typing_extensions import Protocol

from galaxy.util.resources import (
    as_file,
    resource_path,
    Traversable,
)
from galaxy_test.base import api_asserts
from galaxy_test.base.api_util import random_name
from tool_shed_client.schema import (
    BuildSearchIndexResponse,
    Category,
    CreateCategoryRequest,
    CreateRepositoryRequest,
    DetailedRepository,
    from_legacy_install_info,
    GetInstallInfoRequest,
    GetOrderedInstallableRevisionsRequest,
    InstallInfo,
    OrderedInstallableRevisions,
    PaginatedRepositoryIndexResults,
    RepositoriesByCategory,
    Repository,
    RepositoryIndexRequest,
    RepositoryIndexResponse,
    RepositoryMetadata,
    RepositoryPaginatedIndexRequest,
    RepositorySearchRequest,
    RepositorySearchResults,
    RepositoryUpdate,
    RepositoryUpdateRequest,
    ResetMetadataOnRepositoryRequest,
    ResetMetadataOnRepositoryResponse,
    ToolSearchRequest,
    ToolSearchResults,
    UpdateRepositoryRequest,
    Version,
)
from .api_util import (
    ensure_user_with_email,
    ShedApiInteractor,
)

HasRepositoryId = Union[str, Repository]

DEFAULT_PREFIX = "repofortest"
TEST_DATA_REPO_FILES = resource_path(__name__, "../test_data")
COLUMN_MAKER_PATH = TEST_DATA_REPO_FILES.joinpath("column_maker/column_maker.tar")
COLUMN_MAKER_1_1_1_PATH = TEST_DATA_REPO_FILES.joinpath("column_maker/column_maker_1.1.1.tar")
DEFAULT_COMMIT_MESSAGE = "a test commit message"


def repo_files(test_data_path: str) -> Iterator[Path]:
    repos = TEST_DATA_REPO_FILES.joinpath(f"repos/{test_data_path}")
    for child in sorted(_.name for _ in repos.iterdir()):
        with as_file(repos.joinpath(child)) as path:
            yield path


def repo_tars(test_data_path: str) -> Iterator[Path]:
    for path in repo_files(test_data_path):
        assert path.is_dir()
        prefix = f"shedtest_{test_data_path}_{path.name}_"
        with NamedTemporaryFile(prefix=prefix) as tf:
            with tarfile.open(tf.name, "w:gz") as tar:
                tar.add(str(path.absolute()), arcname=test_data_path or path.name)
            yield Path(tf.name)


class HostsTestToolShed(Protocol):
    host: str
    port: Optional[str]


class ToolShedPopulator:
    """Utilities for easy fixture creation of tool shed related things."""

    _admin_api_interactor: ShedApiInteractor
    _api_interactor: ShedApiInteractor

    def __init__(self, admin_api_interactor: ShedApiInteractor, api_interactor: ShedApiInteractor):
        self._admin_api_interactor = admin_api_interactor
        self._api_interactor = api_interactor

    def setup_bismark_repo(
        self,
        repository_id: Optional[HasRepositoryId] = None,
        end: Optional[int] = None,
        category_id: Optional[str] = None,
    ) -> HasRepositoryId:
        if repository_id is None:
            category_id = category_id or self.new_category(prefix="testbismark").id
            repository_id = self.new_repository(category_id, prefix="testbismark")
        return self.setup_test_data_repo_by_id("bismark", repository_id, assert_ok=False, end=end)

    def setup_test_data_repo_by_id(
        self,
        test_data_path: str,
        repository_id: Optional[HasRepositoryId] = None,
        assert_ok=True,
        start: int = 0,
        end: Optional[int] = None,
    ) -> HasRepositoryId:
        if repository_id is None:
            prefix = test_data_path.replace("_", "")
            category_id = self.new_category(prefix=prefix).id
            repository = self.new_repository(category_id, prefix=prefix)
            repository_id = repository.id

        assert repository_id

        for index, repo_tar in enumerate(repo_tars(test_data_path)):
            if index < start:
                continue

            if end and index >= end:
                break

            commit_message = f"Updating {test_data_path} with index {index} with tar {repo_tar}"
            response = self.upload_revision_raw(repository_id, repo_tar, commit_message)
            if assert_ok:
                api_asserts.assert_status_code_is_ok(response)
                assert RepositoryUpdate(root=response.json()).is_ok
        return repository_id

    def setup_test_data_repo(
        self,
        test_data_path: str,
        repository: Optional[Repository] = None,
        assert_ok=True,
        start: int = 0,
        end: Optional[int] = None,
        category_id: Optional[str] = None,
    ) -> Repository:
        if repository is None:
            prefix = test_data_path.replace("_", "")
            if category_id is None:
                category_id = self.new_category(prefix=prefix).id
            repository = self.new_repository(category_id, prefix=prefix)
        self.setup_test_data_repo_by_id(test_data_path, repository, assert_ok=assert_ok, start=start, end=end)
        return repository

    def setup_column_maker_repo(
        self,
        prefix=DEFAULT_PREFIX,
        category_id: Optional[str] = None,
    ) -> Repository:
        if category_id is None:
            category_id = self.new_category(prefix=prefix).id
        assert category_id
        repository = self.new_repository(category_id, prefix=prefix)
        repository_id = repository.id
        assert repository_id

        response = self.upload_revision(
            repository_id,
            COLUMN_MAKER_PATH,
        )
        assert response.is_ok
        return repository

    def setup_column_maker_and_get_metadata(self, prefix=DEFAULT_PREFIX) -> RepositoryMetadata:
        repository = self.setup_column_maker_repo(prefix=prefix)
        return self.get_metadata(repository)

    def get_install_info_for_repository(self, has_repository_id: HasRepositoryId) -> InstallInfo:
        repository_id = self._repository_id(has_repository_id)
        metadata = self.get_metadata(repository_id, True)
        return self.get_install_info(metadata)

    def get_install_info(self, repository_metadata: RepositoryMetadata) -> InstallInfo:
        revision_metadata = repository_metadata.latest_revision
        repo = revision_metadata.repository
        request = GetInstallInfoRequest(
            owner=repo.owner,
            name=repo.name,
            changeset_revision=revision_metadata.changeset_revision,
        )
        revisions_response = self._api_interactor.get(
            "repositories/get_repository_revision_install_info", params=request.model_dump()
        )
        api_asserts.assert_status_code_is_ok(revisions_response)
        return from_legacy_install_info(revisions_response.json())

    def update_column_maker_repo(self, repository: HasRepositoryId) -> RepositoryUpdate:
        response = self.upload_revision(
            repository,
            COLUMN_MAKER_1_1_1_PATH,
        )
        return response

    def upload_revision_raw(
        self, repository: HasRepositoryId, path: Traversable, commit_message: str = DEFAULT_COMMIT_MESSAGE
    ) -> requests.Response:
        body = RepositoryUpdateRequest(
            commit_message=commit_message,
        )
        files = {"file": path.open("rb")}
        repository_id = self._repository_id(repository)
        response = self._api_interactor.post(
            f"repositories/{repository_id}/changeset_revision", params=body.model_dump(), files=files
        )
        return response

    def update_raw(self, repository: HasRepositoryId, request: UpdateRepositoryRequest) -> requests.Response:
        repository_id = self._repository_id(repository)
        body_json = request.model_dump(exclude_unset=True, by_alias=True)
        put_response = self._api_interactor.put(f"repositories/{repository_id}", json=body_json)
        return put_response

    def update(self, repository: HasRepositoryId, request: UpdateRepositoryRequest) -> Repository:
        response = self.update_raw(repository, request)
        api_asserts.assert_status_code_is_ok(response)
        return Repository(**response.json())

    def upload_revision(
        self, repository: HasRepositoryId, path: Traversable, commit_message: str = DEFAULT_COMMIT_MESSAGE
    ) -> RepositoryUpdate:
        response = self.upload_revision_raw(repository, path, commit_message=commit_message)
        if response.status_code != 200:
            response_json = None
            err_msg = None
            try:
                response_json = response.json()
            except Exception:
                pass
            if response_json and "err_msg" in response_json:
                err_msg = response_json["err_msg"]
            if err_msg and "No changes" in err_msg:
                assert_msg = f"Updating repository [{repository}] with path [{path}] and commit_message {commit_message} failed to update repository contents, no changes found. Response: [{response_json}]"
                raise AssertionError(assert_msg)
            api_asserts.assert_status_code_is_ok(response)
        return RepositoryUpdate(root=response.json())

    def new_repository(self, category_ids: Union[list[str], str], prefix: str = DEFAULT_PREFIX) -> Repository:
        name = random_name(prefix=prefix)
        synopsis = random_name(prefix=prefix)
        request = CreateRepositoryRequest(
            name=name,
            synopsis=synopsis,
            category_ids=category_ids,
        )
        return self.create_repository(request)

    def create_repository(self, request: CreateRepositoryRequest) -> Repository:
        response = self._api_interactor.post("repositories", json=request.model_dump(by_alias=True))
        api_asserts.assert_status_code_is_ok(response)
        return Repository(**response.json())

    def reindex(self) -> BuildSearchIndexResponse:
        index_response = self._admin_api_interactor.put("tools/build_search_index")
        index_response.raise_for_status()
        return BuildSearchIndexResponse(**index_response.json())

    def new_category(
        self, name: Optional[str] = None, description: Optional[str] = None, prefix=DEFAULT_PREFIX
    ) -> Category:
        category_name = name or random_name(prefix=prefix)
        category_description = description or "testcreaterepo"
        request = CreateCategoryRequest(name=category_name, description=category_description)
        response = self._admin_api_interactor.post("categories", json=request.model_dump())
        response.raise_for_status()
        return Category(**response.json())

    def get_categories(self) -> list[Category]:
        response = self._api_interactor.get("categories")
        response.raise_for_status()
        return [Category(**c) for c in response.json()]

    def get_category_with_id(self, category_id: str) -> Category:
        response = self._api_interactor.get(f"categories/{category_id}")
        response.raise_for_status()
        return Category(**response.json())

    def get_category_with_name(self, name: str) -> Category:
        categories = [c for c in self.get_categories() if c.name == name]
        if not categories:
            raise ValueError(f"No category with name {name} found.")
        return categories[0]

    def repositories_by_category(self, category_id: str) -> RepositoriesByCategory:
        response = self._api_interactor.get(f"categories/{category_id}/repositories")
        response.raise_for_status()
        return RepositoriesByCategory(**response.json())

    def assert_category_has_n_repositories(self, category_id: str, n: int):
        category_repos = self.repositories_by_category(category_id)
        assert category_repos.repository_count == n
        assert len(category_repos.repositories) == n

    def get_ordered_installable_revisions(self, owner: str, name: str) -> OrderedInstallableRevisions:
        request = GetOrderedInstallableRevisionsRequest(owner=owner, name=name)
        revisions_response = self._api_interactor.get(
            "repositories/get_ordered_installable_revisions", params=request.model_dump()
        )
        api_asserts.assert_status_code_is_ok(revisions_response)
        return OrderedInstallableRevisions(root=revisions_response.json())

    def assert_has_n_installable_revisions(self, repository: Repository, n: int):
        revisions = self.get_ordered_installable_revisions(repository.owner, repository.name)
        actual_n = len(revisions.root)
        assert actual_n == n, f"Expected {n} repository revisions, found {actual_n} for {repository}"

    def get_repository_for(self, owner: str, name: str, deleted: str = "false") -> Optional[Repository]:
        request = RepositoryIndexRequest(
            owner=owner,
            name=name,
            deleted=deleted,
        )
        index = self.repository_index(request)
        return index.root[0] if index.root else None

    def repository_index(self, request: Optional[RepositoryIndexRequest]) -> RepositoryIndexResponse:
        repository_response = self._api_interactor.get("repositories", params=(request.model_dump() if request else {}))
        api_asserts.assert_status_code_is_ok(repository_response)
        return RepositoryIndexResponse(root=repository_response.json())

    def repository_index_paginated(
        self, request: Optional[RepositoryPaginatedIndexRequest]
    ) -> PaginatedRepositoryIndexResults:
        repository_response = self._api_interactor.get(
            "repositories", params=(request.model_dump() if request else {"page": 1})
        )
        api_asserts.assert_status_code_is_ok(repository_response)
        return PaginatedRepositoryIndexResults(**repository_response.json())

    def get_usernames_allowed_to_push(self, repository: HasRepositoryId) -> list[str]:
        repository_id = self._repository_id(repository)
        show_response = self._api_interactor.get(f"repositories/{repository_id}/allow_push")
        show_response.raise_for_status()
        as_list = show_response.json()
        assert isinstance(as_list, list)
        return as_list

    def allow_user_to_push(self, repository: HasRepositoryId, username: str) -> None:
        repository_id = self._repository_id(repository)
        post_response = self._api_interactor.post(f"repositories/{repository_id}/allow_push/{username}")
        post_response.raise_for_status()

    def disallow_user_to_push(self, repository: HasRepositoryId, username: str) -> None:
        repository_id = self._repository_id(repository)
        delete_response = self._api_interactor.delete(f"repositories/{repository_id}/allow_push/{username}")
        delete_response.raise_for_status()

    def set_malicious(self, repository: HasRepositoryId, changeset_revision: str):
        repository_id = self._repository_id(repository)
        put_response = self._api_interactor.put(
            f"repositories/{repository_id}/revisions/{changeset_revision}/malicious"
        )
        put_response.raise_for_status()

    def unset_malicious(self, repository: HasRepositoryId, changeset_revision: str):
        repository_id = self._repository_id(repository)
        delete_response = self._api_interactor.delete(
            f"repositories/{repository_id}/revisions/{changeset_revision}/malicious"
        )
        delete_response.raise_for_status()

    def tip_is_malicious(self, repository: HasRepositoryId) -> bool:
        repository_metadata = self.get_metadata(repository)
        revision = repository_metadata.latest_revision
        return revision.malicious

    def set_deprecated(self, repository: HasRepositoryId):
        repository_id = self._repository_id(repository)
        put_response = self._api_interactor.put(f"repositories/{repository_id}/deprecated")
        put_response.raise_for_status()

    def unset_deprecated(self, repository: HasRepositoryId):
        repository_id = self._repository_id(repository)
        delete_response = self._api_interactor.delete(f"repositories/{repository_id}/deprecated")
        delete_response.raise_for_status()

    def get_repository(self, repository: HasRepositoryId) -> DetailedRepository:
        repository_id = self._repository_id(repository)
        repository_response = self._api_interactor.get(f"repositories/{repository_id}")
        repository_response.raise_for_status()
        return DetailedRepository(**repository_response.json())

    def is_deprecated(self, repository: HasRepositoryId) -> bool:
        return self.get_repository(repository).deprecated

    def get_metadata(self, repository: HasRepositoryId, downloadable_only=True) -> RepositoryMetadata:
        repository_id = self._repository_id(repository)
        metadata_response = self._api_interactor.get(
            f"repositories/{repository_id}/metadata?downloadable_only={downloadable_only}"
        )
        api_asserts.assert_status_code_is_ok(metadata_response)
        return RepositoryMetadata(root=metadata_response.json())

    def reset_metadata(self, repository: HasRepositoryId) -> ResetMetadataOnRepositoryResponse:
        repository_id = self._repository_id(repository)
        request = ResetMetadataOnRepositoryRequest(repository_id=repository_id)
        reset_response = self._api_interactor.post(
            "repositories/reset_metadata_on_repository", json=request.model_dump()
        )
        api_asserts.assert_status_code_is_ok(reset_response)
        return ResetMetadataOnRepositoryResponse(**reset_response.json())

    def version(self) -> Version:
        version_response = self._admin_api_interactor.get("version")
        api_asserts.assert_status_code_is_ok(version_response)
        return Version(**version_response.json())

    def tool_search_query(self, query: str) -> ToolSearchResults:
        return self.tool_search(ToolSearchRequest(q=query))

    def tool_search(self, search_request: ToolSearchRequest) -> ToolSearchResults:
        search_response = self._api_interactor.get("tools", params=search_request.model_dump())
        api_asserts.assert_status_code_is_ok(search_response)
        return ToolSearchResults(**search_response.json())

    def tool_guid(
        self, shed_host: HostsTestToolShed, repository: Repository, tool_id: str, tool_version: Optional[str] = None
    ) -> str:
        owner = repository.owner
        name = repository.name
        port = shed_host.port
        if port in [None, "80", "443"]:
            host_and_port = shed_host.host
        else:
            host_and_port = f"{shed_host.host}:{shed_host.port}"
        tool_id_base = f"{host_and_port}/repos/{owner}/{name}/{tool_id}"
        if tool_version is None:
            return tool_id_base
        else:
            return f"{tool_id_base}/{tool_version}"

    def repo_search_query(self, query: str) -> RepositorySearchResults:
        return self.repo_search(RepositorySearchRequest(q=query))

    def repo_search(self, repo_search_request: RepositorySearchRequest) -> RepositorySearchResults:
        search_response = self._api_interactor.get("repositories", params=repo_search_request.model_dump())
        api_asserts.assert_status_code_is_ok(search_response)
        return RepositorySearchResults(**search_response.json())

    def delete_api_key(self) -> None:
        response = self._api_interactor.delete("users/current/api_key")
        response.raise_for_status()

    def create_new_api_key(self) -> str:
        response = self._api_interactor.post("users/current/api_key")
        response.raise_for_status()
        return response.json()

    def guid(self, repository: Repository, tool_id: str, tool_version: str) -> str:
        url = self._api_interactor.url
        base = url.split("://")[1].split("/")[0]
        return f"{base}/repos/{repository.owner}/{repository.name}/{tool_id}/{tool_version}"

    def new_user(self, username: str, password: str):
        return ensure_user_with_email(self._admin_api_interactor, username, password)

    def _repository_id(self, has_id: HasRepositoryId) -> str:
        if isinstance(has_id, Repository):
            return has_id.id
        else:
            return str(has_id)



import logging

from . import (
    repository_suite_definition,
    tool_dependency_definition,
    unrestricted,
)

log = logging.getLogger(__name__)


class Registry:
    def __init__(self):
        self.repository_types_by_label = {}
        self.repository_types_by_label["unrestricted"] = unrestricted.Unrestricted()
        self.repository_types_by_label["repository_suite_definition"] = (
            repository_suite_definition.RepositorySuiteDefinition()
        )
        self.repository_types_by_label["tool_dependency_definition"] = (
            tool_dependency_definition.ToolDependencyDefinition()
        )

    def get_class_by_label(self, label):
        return self.repository_types_by_label.get(label, None)



from urllib.parse import urljoin

from requests import post

from galaxy_test.base import api_asserts
from tool_shed_client.schema import (
    CreateUserRequest,
    User,
)
from ..base.api import (
    ShedApiTestCase,
    skip_if_api_v1,
)
from ..base.api_util import (
    email_to_username,
    ensure_user_with_email,
    get_admin_api_key,
)


class TestShedUsersApi(ShedApiTestCase):
    def test_create_requires_admin(self):
        url = urljoin(self.url, "api/users")
        response = post(url)
        assert response.status_code == 403

    def test_create_user(self):
        url = urljoin(self.url, "api/users")
        headers = {
            "x-api-key": get_admin_api_key(),
        }
        email = "testcreateuser@bx.psu.edu"
        password = "mycoolpassword123"
        username = email_to_username(email)
        body = {
            "email": email,
            "password": password,
            "username": username,
        }
        request = CreateUserRequest(**body)
        response = post(url, json=request.model_dump(), headers=headers)
        api_asserts.assert_status_code_is_ok(response)
        self._verify_username_password(email, password)
        user = User(**response.json())
        assert user.id
        assert user.username == username

    def test_create_user_interactor(self):
        email = "testcreateuserinteractor@bx.psu.edu"
        password = "mycoolpassword123"
        body = {
            "email": email,
            "password": password,
            "username": email_to_username(email),
        }
        response = self.admin_api_interactor.post("users", json=body)
        api_asserts.assert_status_code_is_ok(response)
        self._verify_username_password(email, password)

    def test_ensure_user_with_email(self):
        email = "testcreateuserensure@bx.psu.edu"
        password = "mycoolpassword123"
        ensure_user_with_email(self.admin_api_interactor, email, password)
        self._verify_username_password(email, password)

    def test_simple_index_and_user(self):
        email = "testindexshow@bx.psu.edu"
        password = "mycoolpassword123"
        ensure_user_with_email(self.admin_api_interactor, email, password)
        user_response = self.admin_api_interactor.get("users")
        api_asserts.assert_status_code_is_ok(user_response)
        users = user_response.json()
        assert isinstance(users, list)
        username = email_to_username(email)
        filtered_users = [u for u in users if u["username"] == username]
        assert len(filtered_users) == 1
        user_id = filtered_users[0]["id"]
        show_response = self.admin_api_interactor.get(f"users/{user_id}")
        api_asserts.assert_status_code_is_ok(show_response)
        assert show_response.json()["username"] == username
        assert show_response.json()["id"] == user_id

    @skip_if_api_v1
    def test_api_key_endpoints(self):
        email = "testindexapi@bx.psu.edu"
        password = "mycoolpassword123"
        ensure_user_with_email(self.admin_api_interactor, email, password)
        api_key = self._verify_username_password(email, password)
        second_try_api_key = self._verify_username_password(email, password)
        assert api_key == second_try_api_key

        user_populator = self.populator_for_key(api_key)
        user_populator.delete_api_key()
        new_api_key = self._verify_username_password(email, password)
        assert api_key != new_api_key

        user_populator = self.populator_for_key(new_api_key)
        another_new_api_key = user_populator.create_new_api_key()
        assert new_api_key != another_new_api_key
        assert new_api_key != api_key

    def _verify_username_password(self, email: str, password: str) -> str:
        return self.api_interactor.create_api_key(email, password)



import os
import re
from functools import wraps
from typing import (
    Any,
    Callable,
    Optional,
)
from urllib.parse import urljoin

import requests
from typing_extensions import Literal

from galaxy_test.base.api_asserts import (
    assert_has_keys,
    assert_status_code_is,
    assert_status_code_is_ok,
)
from galaxy_test.base.api_util import baseauth_headers

DEFAULT_TOOL_SHED_BOOTSTRAP_ADMIN_API_KEY = "TEST1234"
DEFAULT_TOOL_SHED_USER_API_KEY = None


def get_admin_api_key() -> str:
    """Test admin API key to use for functional tests.

    This key should be configured as a admin API key and should be able
    to create additional users and keys.
    """
    for key in ["TOOL_SHED_CONFIG_BOOTSTRAP_ADMIN_API_KEY", "TOOL_SHED_CONFIG_OVERRIDE_BOOTSTRAP_ADMIN_API_KEY"]:
        value = os.environ.get(key, None)
        if value:
            return value
    return DEFAULT_TOOL_SHED_BOOTSTRAP_ADMIN_API_KEY


def get_user_api_key() -> Optional[str]:
    """Test user API key to use for functional tests.

    If set, this should drive API based testing - if not set an admin API key will
    be used to create a new user and API key for tests.
    """
    return os.environ.get("TOOL_SHED_TEST_USER_API_KEY", DEFAULT_TOOL_SHED_USER_API_KEY)


def decorate_method(method: Callable):
    @wraps(method)
    def wrapper(self: "ShedApiInteractor", route: str, **kwd) -> requests.Response:
        url = urljoin(self.url, f"api/{route}")
        kwd = self._append_headers(kwd)
        return method(url, **kwd)

    return wrapper


class ShedApiInteractor:
    url: str
    api_key: str

    def __init__(self, url: str, api_key: str):
        self.url = url
        self.api_key = api_key

    def create_api_key(self, email: str, password: str) -> str:
        headers = baseauth_headers(email, password)
        url = urljoin(self.url, "api/authenticate/baseauth")
        auth_response = requests.get(url, headers=headers)
        assert_status_code_is(auth_response, 200)
        auth_dict = auth_response.json()
        assert_has_keys(auth_dict, "api_key")
        return auth_dict["api_key"]

    def _append_headers(self, kwd):
        if "admin" in kwd:
            key = get_admin_api_key()
        else:
            key = self.api_key

        headers = kwd.get("headers", {})
        headers["x-api-key"] = key
        kwd["headers"] = headers
        return kwd

    get = decorate_method(requests.get)
    post = decorate_method(requests.post)
    put = decorate_method(requests.put)
    delete = decorate_method(requests.delete)

    @property
    def api_version(self) -> Literal["v1", "v2"]:
        config = self.version()
        api_version = config.get("api_version", "v1")
        return api_version

    def version(self) -> dict[str, Any]:
        response = self.get("version")
        response.raise_for_status()
        return response.json()

    @property
    def hg_url_base(self):
        return self.url


def create_user(admin_interactor: ShedApiInteractor, user_dict: dict[str, Any], assert_ok=True) -> dict[str, Any]:
    email = user_dict["email"]
    if "password" not in user_dict:
        user_dict["password"] = "testpass"
    if "remote_user_email" not in user_dict:
        user_dict["remote_user_email"] = email
    response = admin_interactor.post("users", json=user_dict)
    if assert_ok:
        assert_status_code_is_ok(response)
    return response.json()


def ensure_user_with_email(
    admin_api_interactor: ShedApiInteractor, email: str, password: Optional[str]
) -> dict[str, Any]:
    all_users_response = admin_api_interactor.get("users")
    try:
        all_users_response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Failed to verify user with email [{email}] exists - perhaps you're targeting the wrong ToolShed server or using an incorrect admin API key. HTTP error: {e}"
        )
    username = email_to_username(email)
    all_users = all_users_response.json()
    try:
        test_user = [user for user in all_users if user["username"] == username][0]
    except IndexError:
        request = {"email": email, "username": username, "password": password}
        test_user = create_user(admin_api_interactor, request, assert_ok=False)
    return test_user


def email_to_username(email: str) -> str:
    """Pattern used for test user generation - does not use the API."""
    return re.sub(r"[^a-z-\d]", "--", email.lower())



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

repository_name = "bismark_0070"
repository_description = "Galaxy's bismark wrapper"
repository_long_description = "Long description of Galaxy's bismark wrapper"
category_name = "Test 0070 Invalid Tool Revisions"
category_description = "Test 1070 for a repository with an invalid tool."


class TestFreebayesRepository(ShedTwillTestCase):
    """Test repository with multiple revisions with invalid tools."""

    requires_galaxy = True

    def test_0000_create_or_login_admin_user(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_ensure_existence_of_repository_and_category(self):
        """Create freebayes repository and upload only freebayes.xml. This should result in an error message and invalid tool."""
        self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.user_populator().setup_bismark_repo(repository)

    def test_0010_browse_tool_shed(self):
        """Browse the available tool sheds in this Galaxy instance and preview the bismark repository."""
        self.browse_tool_shed(url=self.url, strings_displayed=[category_name])
        category = self.populator.get_category_with_name(category_name)
        self.browse_category(category, strings_displayed=[repository_name])
        self.preview_repository_in_tool_shed(
            repository_name, common.test_user_1_name, strings_displayed=[repository_name]
        )

    def test_0015_install_freebayes_repository(self):
        """Install the test repository without installing tool dependencies."""
        self._install_repository(
            repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            new_tool_panel_section_label="test_1070",
        )
        installed_repository = self._get_installed_repository_by_name_owner(repository_name, common.test_user_1_name)
        assert self._get_installed_repository_for(
            common.test_user_1, repository_name, installed_repository.installed_changeset_revision
        )
        self.update_installed_repository(installed_repository, verify_no_updates=True)
        self._assert_repo_has_invalid_tool_in_file(installed_repository, "bismark_bowtie_wrapper.xml")



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_repository_name = "column_maker_0080"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_0080"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

category_name = "Test 0080 Advanced Circular Dependencies"
category_description = "Test circular dependency features"


class TestRepositoryCircularDependencies(ShedTwillTestCase):
    """Verify that the code correctly handles circular dependencies."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_column_repository(self):
        """Create and populate the column_maker repository."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "column_maker/column_maker.tar",
            commit_message="Uploaded column_maker tarball.",
        )

    def test_0005_create_convert_repository(self):
        """Create and populate the convert_chars repository."""
        self.login(email=common.admin_email, username=common.admin_username)
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=convert_repository_name,
            description=convert_repository_description,
            long_description=convert_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "convert_chars/convert_chars.tar",
            commit_message="Uploaded convert_chars tarball.",
        )

    def test_0020_create_repository_dependencies(self):
        """Upload a repository_dependencies.xml file that specifies the current revision of convert_chars_0080 to the column_maker_0080 repository."""
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        repository_dependencies_path = self.generate_temp_path("test_0080", additional_paths=["convert"])
        repository_tuple = (
            self.url,
            convert_repository.name,
            convert_repository.owner,
            self.get_repository_tip(convert_repository),
        )
        self.create_repository_dependency(
            repository=column_repository, repository_tuples=[repository_tuple], filepath=repository_dependencies_path
        )

    def test_0025_create_dependency_on_filtering(self):
        """Upload a repository_dependencies.xml file that specifies the current revision of filtering to the freebayes_0040 repository."""
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        repository_dependencies_path = self.generate_temp_path("test_0080", additional_paths=["convert"])
        repository_tuple = (
            self.url,
            column_repository.name,
            column_repository.owner,
            self.get_repository_tip(column_repository),
        )
        self.create_repository_dependency(
            repository=convert_repository, repository_tuples=[repository_tuple], filepath=repository_dependencies_path
        )

    def test_0030_verify_repository_dependencies(self):
        """Verify that each repository can depend on the other without causing an infinite loop."""
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        self.check_repository_dependency(
            convert_repository, column_repository, self.get_repository_tip(column_repository)
        )
        self.check_repository_dependency(
            column_repository, convert_repository, self.get_repository_tip(convert_repository)
        )

    def test_0035_verify_repository_metadata(self):
        """Verify that resetting the metadata does not change it."""
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        for repository in [column_repository, convert_repository]:
            self.verify_unchanged_repository_metadata(repository)



from markupsafe import escape as raw_escape

ALLOWED_ELEMENTS = ["<b>", "</b>", "<br/>"]
ALLOWED_MAP = {x: raw_escape(x) for x in ALLOWED_ELEMENTS}


def escape(string):
    """A tool shed variant of markupsafe.escape that allows a select few
    HTML elements that are repeatedly used in messages created deep
    in the toolshed components. Ideally abstract things would be produced
    in these components and messages in the views or client side - this is
    what should be worked toward - but for now - we have this hack.

    >>> assert escape(u"A <b>cómplǐcḁtëd strĩñg</b>") == u'A <b>cómplǐcḁtëd strĩñg</b>'
    """
    escaped = str(raw_escape(string))
    # Unescape few selected tags.
    for key, value in ALLOWED_MAP.items():
        escaped = escaped.replace(value, key)
    return escaped



import tempfile

import twill.commands as tc
from twill.browser import FormElement  # type:ignore[attr-defined, unused-ignore]

from galaxy.util import smart_str
from .browser import (
    FormValueType,
    ShedBrowser,
)

tc.options["equiv_refresh_interval"] = 0
# Resetting all repository metadata can take a really long time
tc.timeout(240)


def visit_url(url: str, allowed_codes: list[int]) -> str:
    tc.go(url)
    return_code = tc.browser.code
    assert return_code in allowed_codes, "Invalid HTTP return code {}, allowed codes: {}".format(
        return_code,
        ", ".join(str(code) for code in allowed_codes),
    )
    return url


def page_content() -> str:
    return tc.browser.html


class TwillShedBrowser(ShedBrowser):
    def visit_url(self, url: str, allowed_codes: list[int]) -> str:
        return visit_url(url, allowed_codes=allowed_codes)

    def page_content(self) -> str:
        """
        Return the last visited page (usually HTML, but can binary data as
        well).
        """
        return page_content()

    def check_page_for_string(self, patt: str) -> None:
        page = self.page_content()
        if page.find(patt) == -1:
            fname = self.write_temp_file(page)
            errmsg = f"no match to '{patt}'\npage content written to '{fname}'\npage: [[{page}]]"
            raise AssertionError(errmsg)

    def check_string_not_in_page(self, patt: str) -> None:
        page = self.page_content()
        if page.find(patt) != -1:
            fname = self.write_temp_file(page)
            errmsg = f"string ({patt}) incorrectly displayed in page.\npage content written to '{fname}'"
            raise AssertionError(errmsg)

    def write_temp_file(self, content, suffix=".html"):
        with tempfile.NamedTemporaryFile(suffix=suffix, prefix="twilltestcase-", delete=False) as fh:
            fh.write(smart_str(content))
        return fh.name

    def submit_form_with_name(self, form_name: str, button="runtool_btn", **kwd):
        forms_by_name: dict[str, FormElement] = {f.get("name"): f for f in self._show_forms()}
        form = forms_by_name[form_name]
        self._submit_form(form, button, **kwd)

    def _show_forms(self) -> list[FormElement]:
        """Shows form, helpful for debugging new tests"""
        return tc.browser.forms

    def submit_form(self, form_no=-1, button="runtool_btn", form=None, **kwd):
        if form is None:
            try:
                form = self._show_forms()[form_no]
            except IndexError:
                raise ValueError("No form to submit found")
        self._submit_form(form, button, **kwd)

    def _submit_form(self, form, button, **kwd):
        controls = {c.name: c for c in form.inputs}
        form_name = form.get("name")
        for control_name, control_value in kwd.items():
            if control_name not in controls:
                continue  # these cannot be handled safely - cause the test to barf out
            if not isinstance(control_value, list):
                control_value = [str(control_value)]
            control = controls[control_name]
            control_type = getattr(control, "type", None)
            if control_type in (
                "text",
                "textfield",
                "submit",
                "password",
                "TextareaElement",
                "checkbox",
                "radio",
                None,
            ):
                for cv in control_value:
                    tc.fv(form_name, control.name, cv)
            else:
                # Add conditions for other control types here when necessary.
                pass
        tc.submit(button)

    def fill_form_value(self, form_name: str, control_name: str, value: FormValueType):
        value = str(value)
        tc.fv(form_name, control_name, value)

    def edit_repository_categories(self, categories_to_add: list[str], categories_to_remove: list[str]) -> None:
        """Select some new categories and then restore the component."""
        strings_displayed = []
        strings_not_displayed = []
        for category in categories_to_add:
            self.fill_form_value("categories", "category_id", f"+{category}")
            strings_displayed.append(f"selected>{category}")
        for category in categories_to_remove:
            self.fill_form_value("categories", "category_id", f"-{category}")
            strings_not_displayed.append(f"selected>{category}")
        self.submit_form_with_name("categories", "manage_categories_button")
        self._check_for_strings(strings_displayed, strings_not_displayed)

        strings_displayed = []
        strings_not_displayed = []
        for category in categories_to_remove:
            self.fill_form_value("categories", "category_id", f"+{category}")
            strings_displayed.append(f"selected>{category}")
        for category in categories_to_add:
            self.fill_form_value("categories", "category_id", f"-{category}")
            strings_not_displayed.append(f"selected>{category}")
        self.submit_form_with_name("categories", "manage_categories_button")
        self._check_for_strings(strings_displayed, strings_not_displayed)

    def grant_users_access(self, usernames: list[str]):
        for username in usernames:
            self.fill_form_value("user_access", "allow_push", f"+{username}")
        self.submit_form_with_name("user_access", "user_access_button")

    @property
    def is_twill(self) -> bool:
        return True

    def _check_for_strings(self, strings_displayed: list[str], strings_not_displayed: list[str]):
        if strings_displayed:
            for check_str in strings_displayed:
                self.check_page_for_string(check_str)
        if strings_not_displayed:
            for check_str in strings_not_displayed:
                self.check_string_not_in_page(check_str)



import logging

import tool_shed.repository_types.util as rt_util
from galaxy.util import unicodify
from tool_shed.repository_types.metadata import TipOnly
from tool_shed.util import basic_util

log = logging.getLogger(__name__)


class RepositorySuiteDefinition(TipOnly):
    def __init__(self):
        self.type = rt_util.REPOSITORY_SUITE_DEFINITION
        self.label = "Repository suite definition"
        self.valid_file_names = ["repository_dependencies.xml"]

    def is_valid_for_type(self, repository, revisions_to_check=None):
        """
        Inspect the received repository's contents to determine if they abide by the rules defined for
        the contents of this type.  If the received revisions_to_check is a list of changeset revisions,
        then inspection will be restricted to the revisions in the list.
        """
        repo = repository.hg_repo
        if revisions_to_check:
            changeset_revisions = revisions_to_check
        else:
            changeset_revisions = repo.changelog
        for changeset in changeset_revisions:
            ctx = repo[changeset]
            # Inspect all files in the changeset (in sorted order) to make sure there is only one and it
            # is named repository_dependencies.xml.
            files_changed_in_changeset = ctx.files()
            for file_path in files_changed_in_changeset:
                file_name = basic_util.strip_path(unicodify(file_path))
                if file_name not in self.valid_file_names:
                    return False
        return True



import os
from typing import (
    Optional,
)

from galaxy.model.migrations.base import pop_arg_from_args
from galaxy.util.properties import (
    find_config_file,
    get_data_dir,
    load_app_properties,
)

DEFAULT_CONFIG_NAMES = ["tool_shed", "tool_shed_wsgi"]
CONFIG_FILE_ARG = "--toolshed-config"
CONFIG_DIR_NAME = "config"
TOOLSHED_CONFIG_PREFIX = "TOOL_SHED_CONFIG_"


def get_dburl(argv: list[str], cwd: str) -> str:
    """Return db url."""
    config_file = pop_arg_from_args(argv, CONFIG_FILE_ARG)
    return get_dburl_from_file(cwd, config_file)


def get_dburl_from_file(cwd: str, config_file: Optional[str] = None) -> str:
    if config_file is None:
        cwds = [cwd, os.path.join(cwd, CONFIG_DIR_NAME)]
        config_file = find_config_file(DEFAULT_CONFIG_NAMES, dirs=cwds)

    properties = load_app_properties(
        config_file=config_file, config_prefix=TOOLSHED_CONFIG_PREFIX, config_section="tool_shed"
    )
    default_url = f"sqlite:///{os.path.join(get_data_dir(properties), 'community.sqlite')}?isolation_level=IMMEDIATE"
    url = properties.get("database_connection", default_url)
    return url



from typing import TYPE_CHECKING

from galaxy.structured_app import BasicSharedApp

if TYPE_CHECKING:
    from galaxy.tools.data import ToolDataTableManager
    from tool_shed.managers.model_cache import ModelCache
    from tool_shed.repository_registry import RegistryInterface
    from tool_shed.repository_types.registry import Registry as RepositoryTypesRegistry
    from tool_shed.util.hgweb_config import HgWebConfigManager
    from tool_shed.webapp.model import mapping
    from tool_shed.webapp.security import CommunityRBACAgent


class ToolShedApp(BasicSharedApp):
    repository_types_registry: "RepositoryTypesRegistry"
    model: "mapping.ToolShedModelMapping"
    repository_registry: "RegistryInterface"
    hgweb_config_manager: "HgWebConfigManager"
    security_agent: "CommunityRBACAgent"
    model_cache: "ModelCache"
    tool_data_tables: "ToolDataTableManager"



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

category_name = "Test 0460 Automatic repository revision completion"
category_description = "Test 0460 Automatic repository revision completion"
bwa_repository_name = "package_bwa_0_5_9_0460"
bwa_repository_description = (
    "Contains a tool dependency definition that downloads and compiles version 0.5.9 of the BWA package"
)
bwa_repository_long_description = "bwa (alignment via Burrows-Wheeler transformation) 0.5.9-r16 by Heng Li"

"""
For all steps, verify that the generated dependency points to the tip of the specified repository.

1)  Create category...

2)  Create and populate package_bwa_0_5_9_0460

3)  Create complex_dependency_test_1_0460, complex_dependency_test_2_0460, complex_dependency_test_3_0460,
    complex_dependency_test_4_0460, complex_dependency_test_5_0460.

4)  Upload an uncompressed tool_dependencies.xml to complex_dependency_test_1_0460 that specifies a complex
    repository dependency on package_bwa_0_5_9_0460 without a specified changeset revision or tool shed url.

5)  Upload a tarball to complex_dependency_test_2_0460 with a tool_dependencies.xml in the root of the tarball.

6)  Upload a tarball to complex_dependency_test_3_0460 with a tool_dependencies.xml in a subfolder within the tarball.

7)  Create hg_tool_dependency_0460 and hg_subfolder_tool_dependency_0460 and populate with tool dependencies.

8)  Upload to complex_dependency_test_4_0460 using the url hg://<tool shed url>/repos/user1/hg_tool_dependency_0460.

9)  Upload to complex_dependency_test_5_0460 using the url hg://<tool shed url>/repos/user1/hg_subfolder_tool_dependency_0460.

10) Create repository_dependency_test_1_0460, repository_dependency_test_2_0460, repository_dependency_test_3_0460,
    repository_dependency_test_4_0460, repository_dependency_test_4_0460.

11) Upload an uncompressed repository_dependencies.xml to repository_dependency_test_1_0460 that specifies a
    repository dependency on the package without a specified changeset revision or tool shed url.

12) Upload a tarball to repository_dependency_test_1_0460 with a repository_dependencies.xml in the root of the tarball.

13) Upload a tarball to repository_dependency_test_1_0460 with a repository_dependencies.xml in a subfolder within the tarball.

14) Create hg_repository_dependency_0460 and populate with repository_dependencies.xml.

15) Upload to repository_dependency_test_4_0460 using the url hg://<tool shed url>/repos/user1/hg_repository_dependency_0460.

16) Upload to repository_dependency_test_5_0460 using the url hg://<tool shed url>/repos/user1/hg_repository_dependency_0460.
"""


class TestAutomaticDependencyRevision(ShedTwillTestCase):
    """Test defining repository dependencies without specifying the changeset revision."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_datatypes_repository(self):
        """Create and populate the emboss_datatypes_0460 repository

        This is step 1 - Create and populate emboss_datatypes_0460.
        """
        self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)

    def test_0010_create_bwa_package_repository(self):
        """Create and populate the package_bwa_0_5_9_0460 repository.

        This is step 2 - Create and populate package_bwa_0_5_9_0460.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=bwa_repository_name,
            description=bwa_repository_description,
            long_description=bwa_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.add_file_to_repository(
            repository,
            "bwa/complex/tool_dependencies.xml",
            commit_message="Populate package_bwa_0_5_9_0460 with a tool dependency definition.",
        )

    def test_0015_create_tool_dependency_repositories(self):
        """Create repositories for testing complex dependency generation.

        This is step 3 - Create complex_dependency_test_1_0460, complex_dependency_test_2_0460, complex_dependency_test_3_0460,
        complex_dependency_test_4_0460, complex_dependency_test_5_0460. Each of these repositories will be populated in a way
        that tests a different way to achieve the same resulting dependency structure using complex tool dependencies.
        The different methods being tested are:
        - Upload an uncompressed tool_dependencies.xml to the root of the repository.
        - Upload a tool_dependencies.xml in a tarball, not in a subfolder.
        - Upload a tool_dependencies.xml in a subfolder within a tarball.
        - Upload via url, with the tool_dependencies.xml in the root of another repository.
        - Upload via url, with the tool_dependencies.xml in a subfolder within another repository.
        """
        category = self.populator.get_category_with_name(category_name)
        repository_base_name = "complex_dependency_test_%d_0460"
        repository_base_description = "Test #%d for complex repository dependency definitions."
        repository_base_long_description = "Test #%d for complex repository dependency definitions."
        for number in range(1, 6):
            self.get_or_create_repository(
                name=repository_base_name % number,
                description=repository_base_description % number,
                long_description=repository_base_long_description % number,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )

    def test_0020_populate_complex_dependency_test_1_0460(self):
        """Populate complex_dependency_test_1_0460.

        This is step 4 - Upload an uncompressed tool_dependencies.xml to complex_dependency_test_1_0460 that specifies
        a complex repository dependency on package_bwa_0_5_9_0460 without a specified changeset revision or tool shed url.
        """
        repository = self._get_repository_by_name_and_owner("complex_dependency_test_1_0460", common.test_user_1_name)
        package_repository = self._get_repository_by_name_and_owner("package_bwa_0_5_9_0460", common.test_user_1_name)
        self.add_file_to_repository(repository, "0460_files/tool_dependencies.xml")
        if not self.is_v2:
            changeset_revision = self.get_repository_tip(package_repository)
            strings_displayed = ["package_bwa_0_5_9_0460", "bwa", "0.5.9", "package", changeset_revision]
            self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
            self.display_repository_file_contents(
                repository, filename="tool_dependencies.xml", strings_displayed=[changeset_revision]
            )

    def test_0025_populate_complex_dependency_test_2_0460(self):
        """Populate complex_dependency_test_2_0460.

        This is step 5 - Upload an tarball with tool_dependencies.xml to complex_dependency_test_2_0460 that specifies
        a complex repository dependency on package_bwa_0_5_9_0460 without a specified changeset revision or tool shed url.
        """
        repository = self._get_repository_by_name_and_owner("complex_dependency_test_2_0460", common.test_user_1_name)
        package_repository = self._get_repository_by_name_and_owner("package_bwa_0_5_9_0460", common.test_user_1_name)
        self.commit_tar_to_repository(
            repository,
            "0460_files/tool_dependencies_in_root.tar",
            commit_message="Uploaded complex repository dependency definition.",
        )
        if not self.is_v2:
            changeset_revision = self.get_repository_tip(package_repository)
            strings_displayed = ["package_bwa_0_5_9_0460", "bwa", "0.5.9", "package", changeset_revision]
            self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
            self.display_repository_file_contents(
                repository, filename="tool_dependencies.xml", strings_displayed=[changeset_revision]
            )

    def test_0030_populate_complex_dependency_test_3_0460(self):
        """Populate complex_dependency_test_3_0460.

        This is step 6 - Upload an tarball with tool_dependencies.xml in a subfolder to complex_dependency_test_3_0460 that
        specifies a complex repository dependency on package_bwa_0_5_9_0460 without a specified changeset revision or tool shed url.
        """
        repository = self._get_repository_by_name_and_owner("complex_dependency_test_3_0460", common.test_user_1_name)
        package_repository = self._get_repository_by_name_and_owner("package_bwa_0_5_9_0460", common.test_user_1_name)
        self.commit_tar_to_repository(
            repository,
            "0460_files/tool_dependencies_in_subfolder.tar",
            commit_message="Uploaded complex repository dependency definition.",
        )
        changeset_revision = self.get_repository_tip(package_repository)
        if not self.is_v2:
            strings_displayed = ["package_bwa_0_5_9_0460", "bwa", "0.5.9", "package", changeset_revision]
            self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
            self.display_repository_file_contents(
                repository,
                filename="tool_dependencies.xml",
                filepath="subfolder",
                strings_displayed=[changeset_revision],
            )

    def test_0035_create_repositories_for_url_upload(self):
        """Create and populate hg_tool_dependency_0460 and hg_subfolder_tool_dependency_0460.

        This is step 7 - Create hg_tool_dependency_0460 and hg_subfolder_tool_dependency_0460 and populate with tool dependencies.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name="hg_tool_dependency_0460",
            description=bwa_repository_description,
            long_description=bwa_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.add_file_to_repository(repository, "0460_files/tool_dependencies.xml")
        repository = self.get_or_create_repository(
            name="hg_subfolder_tool_dependency_0460",
            description=bwa_repository_description,
            long_description=bwa_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "0460_files/tool_dependencies_in_subfolder.tar",
            commit_message="Populate hg_subfolder_tool_dependency_0460 with a tool dependency definition.",
        )

    def test_0050_create_repositories_for_simple_dependencies(self):
        """Create repositories for testing simple dependency generation.

        This is step 10 - Create repository_dependency_test_1_0460, repository_dependency_test_2_0460, repository_dependency_test_3_0460,
        repository_dependency_test_4_0460, repository_dependency_test_4_0460.. Each of these repositories will be populated in a way
        that tests a different way to achieve the same resulting dependency structure using complex tool dependencies.
        The different methods being tested are:
        - Upload an uncompressed repository_dependencies.xml to the root of the repository.
        - Upload a repository_dependencies.xml in a tarball, not in a subfolder.
        - Upload a repository_dependencies.xml in a subfolder within a tarball.
        - Upload via url, with the repository_dependencies.xml in the root of another repository.
        - Upload via url, with the repository_dependencies.xml in a subfolder within another repository.
        """
        category = self.populator.get_category_with_name(category_name)
        repository_base_name = "repository_dependency_test_%d_0460"
        repository_base_description = "Test #%d for repository dependency definitions."
        repository_base_long_description = "Test #%d for repository dependency definitions."
        for number in range(1, 6):
            self.get_or_create_repository(
                name=repository_base_name % number,
                description=repository_base_description % number,
                long_description=repository_base_long_description % number,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )

    def test_0055_populate_repository_dependency_test_1_0460(self):
        """Populate repository_dependency_test_1_0460.

        This is step 11 - Upload an uncompressed repository_dependencies.xml to repository_dependency_test_1_0460 that specifies a
        repository dependency on emboss_datatypes_0460 without a specified changeset revision or tool shed url.
        """
        repository = self._get_repository_by_name_and_owner(
            "repository_dependency_test_1_0460", common.test_user_1_name
        )
        package_repository = self._get_repository_by_name_and_owner(bwa_repository_name, common.test_user_1_name)
        self.add_file_to_repository(repository, "0460_files/repository_dependencies.xml")
        changeset_revision = self.get_repository_tip(package_repository)
        if not self.is_v2:
            strings_displayed = [bwa_repository_name, "user1", changeset_revision]
            self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
            self.display_repository_file_contents(
                repository, filename="repository_dependencies.xml", strings_displayed=[changeset_revision]
            )

    def test_0060_populate_repository_dependency_test_2_0460(self):
        """Populate repository_dependency_test_2_0460.

        This is step 12 - Upload a tarball to repository_dependency_test_2_0460 with a repository_dependencies.xml in the root of the tarball.
        """
        repository = self._get_repository_by_name_and_owner(
            "repository_dependency_test_2_0460", common.test_user_1_name
        )
        package_repository = self._get_repository_by_name_and_owner(bwa_repository_name, common.test_user_1_name)
        self.commit_tar_to_repository(
            repository,
            "0460_files/in_root/repository_dependencies_in_root.tar",
            commit_message="Uploaded complex repository dependency definition.",
        )
        changeset_revision = self.get_repository_tip(package_repository)
        if not self.is_v2:
            strings_displayed = [bwa_repository_name, "user1", changeset_revision]
            self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
            self.display_repository_file_contents(
                repository, filename="repository_dependencies.xml", strings_displayed=[changeset_revision]
            )

    def test_0065_populate_repository_dependency_test_3_0460(self):
        """Populate repository_dependency_test_3_0460.

        This is step 13 - Upload a tarball to repository_dependency_test_3_0460 with a repository_dependencies.xml in a
        subfolder within the tarball.
        """
        repository = self._get_repository_by_name_and_owner(
            "repository_dependency_test_3_0460", common.test_user_1_name
        )
        package_repository = self._get_repository_by_name_and_owner(bwa_repository_name, common.test_user_1_name)
        self.commit_tar_to_repository(
            repository,
            "0460_files/in_subfolder/repository_dependencies_in_subfolder.tar",
            commit_message="Uploaded complex repository dependency definition.",
        )
        changeset_revision = self.get_repository_tip(package_repository)
        if not self.is_v2:
            strings_displayed = [bwa_repository_name, "user1", changeset_revision]
            self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
            self.display_repository_file_contents(
                repository,
                filename="repository_dependencies.xml",
                filepath="subfolder",
                strings_displayed=[changeset_revision],
            )

    def test_0070_create_repositories_for_url_upload(self):
        """Create and populate hg_repository_dependency_0460 and hg_subfolder_repository_dependency_0460.

        This is step 14 - Create hg_repository_dependency_0460 and hg_subfolder_repository_dependency_0460 and populate
        with repository dependencies.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name="hg_repository_dependency_0460",
            description=bwa_repository_description,
            long_description=bwa_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.add_file_to_repository(repository, "0460_files/repository_dependencies.xml")
        repository = self.get_or_create_repository(
            name="hg_subfolder_repository_dependency_0460",
            description=bwa_repository_description,
            long_description=bwa_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "0460_files/in_subfolder/repository_dependencies_in_subfolder.tar",
            commit_message="Populate hg_subfolder_repository_dependency_0460 with a tool dependency definition.",
        )



from typing import (
    Any,
    Callable,
)

from sqlalchemy import select

import tool_shed.util.shed_util_common as suc
from galaxy import (
    exceptions,
    web,
)
from tool_shed.context import ProvidesUserContext
from tool_shed.structured_app import ToolShedApp
from tool_shed.webapp.model import Category
from tool_shed_client.schema import (
    Category as CategoryResponse,
    CreateCategoryRequest,
)


class CategoryManager:
    def __init__(self, app: ToolShedApp):
        self.app = app

    def get(self, encoded_category_id: str) -> Category:
        return suc.get_category(self.app, encoded_category_id)

    def create(self, trans: ProvidesUserContext, category_request: CreateCategoryRequest) -> Category:
        name = category_request.name
        description = category_request.description or name
        if name:
            if suc.get_category_by_name(self.app, name):
                raise exceptions.Conflict("A category with that name already exists.")
            else:
                # Create the category
                category = Category(name=name, description=description)
                trans.sa_session.add(category)
                trans.sa_session.commit()
            return category
        else:
            raise exceptions.RequestParameterMissingException('Missing required parameter "name".')

    def index_db(self, trans: ProvidesUserContext, deleted: bool) -> list[Category]:
        if deleted and not trans.user_is_admin:
            raise exceptions.AdminRequiredException("Only administrators can query deleted categories.")
        return list(get_categories_by_deleted(trans.sa_session, deleted))

    def index(self, trans: ProvidesUserContext, deleted: bool) -> list[dict[str, Any]]:
        category_dicts: list[dict[str, Any]] = []
        for category in self.index_db(trans, deleted):
            category_dict = self.to_dict(category)
            category_dicts.append(category_dict)
        return category_dicts

    def to_dict(self, category: Category) -> dict[str, Any]:
        category_dict = category.to_dict(view="collection", value_mapper=get_value_mapper(self.app))
        category_dict["repositories"] = category.active_repository_count()
        category_dict["url"] = web.url_for(
            controller="categories", action="show", id=self.app.security.encode_id(category.id)
        )
        return category_dict

    def to_model(self, category: Category) -> CategoryResponse:
        as_dict = self.to_dict(category)
        return CategoryResponse(
            id=as_dict["id"],
            name=as_dict["name"],
            description=as_dict["description"],
            repositories=as_dict["repositories"],
            deleted=as_dict["deleted"],
        )


def get_value_mapper(app: ToolShedApp) -> dict[str, Callable]:
    value_mapper = {"id": app.security.encode_id}
    return value_mapper


def get_categories_by_deleted(session, deleted):
    stmt = select(Category).where(Category.deleted == deleted).order_by(Category.name)
    return session.scalars(stmt)



import logging

from ..base import common
from ..base.api import skip_if_api_v2
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

emboss_repository_name = "emboss_0430"
emboss_repository_description = "EMBOSS tools for test 0430"
emboss_repository_long_description = "Long description of EMBOSS tools for test 0430"

freebayes_repository_name = "freebayes_0430"
freebayes_repository_description = "Freebayes tool for test 0430"
freebayes_repository_long_description = "Long description of Freebayes tool for test 0430"
"""
1. Create and populate repositories.
2. Browse Custom Datatypes.
3. Browse Tools.
4. Browse Repository Dependencies.
5. Browse Tool Dependencies.
"""


class TestToolShedBrowseUtilities(ShedTwillTestCase):
    """Test browsing for Galaxy utilities."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user.

        Create all the user accounts that are needed for this test script to run independently of other tests.
        Previously created accounts will not be re-created.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0010_create_emboss_repository(self):
        """Create and populate the emboss_0430 repository

        We are at step 1.
        Create the emboss_0430 repository, and populate it with tools.
        """
        category = self.create_category(
            name="Test 0430 Galaxy Utilities", description="Description of Test 0430 Galaxy Utilities category"
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        strings_displayed = self.expect_repo_created_strings(emboss_repository_name)
        emboss_repository = self.get_or_create_repository(
            name=emboss_repository_name,
            description=emboss_repository_description,
            long_description=emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=strings_displayed,
        )
        self.commit_tar_to_repository(
            emboss_repository,
            "emboss/emboss.tar",
            commit_message="Uploaded emboss.tar.",
        )

    def test_0020_create_tool_dependency_repository(self):
        """Create and populate the freebayes_0430 repository

        We are at step 1.
        Create and populate the repository that will have a tool dependency defined.
        """
        category = self.create_category(
            name="Test 0430 Galaxy Utilities", description="Description of Test 0430 Galaxy Utilities category"
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        strings_displayed = self.expect_repo_created_strings(freebayes_repository_name)
        repository = self.get_or_create_repository(
            name=freebayes_repository_name,
            description=freebayes_repository_description,
            long_description=freebayes_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=strings_displayed,
        )
        self.commit_tar_to_repository(
            repository,
            "freebayes/freebayes.tar",
            commit_message="Uploaded freebayes.tar.",
        )

    @skip_if_api_v2
    def test_0030_browse_tools(self):
        """Load the page to browse tools.

        We are at step 3.
        Verify the existence of emboss tools in the browse tools page.
        """
        repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        changeset_revision = self.get_repository_tip(repository)
        strings_displayed = ["EMBOSS", "antigenic1", "5.0.0", changeset_revision, "user1", "emboss_0430"]
        self.browse_tools(strings_displayed=strings_displayed)

    @skip_if_api_v2
    def test_0040_browse_tool_dependencies(self):
        """Browse tool dependencies and look for the right versions of freebayes and samtools.

        We are at step 4.
        Verify that the browse tool dependencies page shows the correct dependencies defined for freebayes_0430.
        """
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        freebayes_changeset_revision = self.get_repository_tip(freebayes_repository)
        strings_displayed = [
            freebayes_changeset_revision,
            "freebayes_0430",
            "user1",
            "0.9.4_9696d0ce8a96",
            "freebayes",
            "samtools",
            "0.1.18",
        ]
        self.browse_tool_dependencies(strings_displayed=strings_displayed)
